# (c) 2016, Matt Martz <matt@sivel.net>
# (c) 2017, Red Hat, Inc.
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

# Copy of github.com/ansible/ansible/lib/ansible/plugins/callback/json.py
# We need to run as a secondary callback not a stdout and we need to control
# the output file location via a zuul environment variable similar to how we
# do in zuul_stream.
# Subclassing wreaks havoc on the module loader and namepsaces
from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import copy
import datetime
import json
import os

from ansible.plugins.loader import PluginLoader
from ansible.plugins.callback import CallbackBase
try:
    # It's here in 2.3
    from ansible.vars import strip_internal_keys
except ImportError:
    try:
        # It's here in 2.4
        from ansible.vars.manager import strip_internal_keys
    except ImportError:
        # It's here in 2.5
        from ansible.vars.clean import strip_internal_keys

from zuul.ansible import logconfig


def current_time():
    return '%sZ' % datetime.datetime.utcnow().isoformat()


class CallbackModule(CallbackBase):
    CALLBACK_VERSION = 2.0
    # aggregate means we can be loaded and not be the stdout plugin
    CALLBACK_TYPE = 'aggregate'
    CALLBACK_NAME = 'zuul_json'

    def __init__(self, display=None):
        super(CallbackModule, self).__init__(display)
        self.results = []
        self.playbook = {}
        logging_config = logconfig.load_job_config(
            os.environ['ZUUL_JOB_LOG_CONFIG'])

        self.output_path = os.path.splitext(
            logging_config.job_output_file)[0] + '.json'

        self._playbook_name = None

    def _new_playbook(self, play):
        extra_vars = play._variable_manager._extra_vars
        self._playbook_name = None

        # TODO(mordred) For now, protect specific variable lookups to make it
        # not absurdly strange to run local tests with the callback plugin
        # enabled. Remove once we have a "run playbook like zuul runs playbook"
        # tool.
        phase = extra_vars.get('zuul_execution_phase')
        index = extra_vars.get('zuul_execution_phase_index')
        playbook = extra_vars.get('zuul_execution_canonical_name_and_path')
        trusted = extra_vars.get('zuul_execution_trusted')
        trusted = True if trusted == "True" else False
        branch = extra_vars.get('zuul_execution_branch')

        self.playbook['playbook'] = playbook
        self.playbook['phase'] = phase
        self.playbook['index'] = index
        self.playbook['trusted'] = trusted
        self.playbook['branch'] = branch

    def _new_play(self, play):
        return {
            'play': {
                'name': play.name,
                'id': str(play._uuid),
                'duration': {
                    'start': current_time()
                }
            },
            'tasks': []
        }

    def _new_task(self, task):
        data = {
            'task': {
                'name': task.name,
                'id': str(task._uuid),
                'duration': {
                    'start': current_time()
                }
            },
            'hosts': {}
        }
        if task._role:
            data['role'] = {
                'name': task._role.get_name(),
                'id': str(task._role._uuid),
                'path': task._role._role_path,
            }
        return data

    def v2_playbook_on_start(self, playbook):
        self._playbook_name = os.path.splitext(playbook._file_name)[0]

    def v2_playbook_on_play_start(self, play):
        if self._playbook_name:
            self._new_playbook(play)

        self.results.append(self._new_play(play))

    def v2_playbook_on_task_start(self, task, is_conditional):
        self.results[-1]['tasks'].append(self._new_task(task))

    def v2_playbook_on_handler_task_start(self, task):
        self.v2_playbook_on_task_start(task, False)

    def v2_runner_on_ok(self, result, **kwargs):
        host = result._host
        action = result._task.action
        # strip_internal_keys makes a deep copy of dict items, but
        # not lists, so we need to create our own complete deep
        # copy first so we don't modify the original.
        myresult = copy.deepcopy(result._result)
        clean_result = strip_internal_keys(myresult)
        self.results[-1]['tasks'][-1]['hosts'][host.name] = clean_result
        end_time = current_time()
        self.results[-1]['tasks'][-1]['task']['duration']['end'] = end_time
        self.results[-1]['play']['duration']['end'] = end_time
        self.results[-1]['tasks'][-1]['hosts'][host.name]['action'] = action

    def v2_runner_on_failed(self, result, **kwargs):
        self.v2_runner_on_ok(result, **kwargs)
        self.results[-1]['tasks'][-1]['hosts'][result._host.name].\
            setdefault('failed', True)

    def v2_runner_on_skipped(self, result, **kwargs):
        self.v2_runner_on_ok(result, **kwargs)
        self.results[-1]['tasks'][-1]['hosts'][result._host.name].\
            setdefault('skipped', True)

    def v2_playbook_on_stats(self, stats):
        """Display info about playbook statistics"""
        hosts = sorted(stats.processed.keys())

        summary = {}
        for h in hosts:
            s = stats.summarize(h)
            summary[h] = s

        self.playbook['plays'] = self.results
        self.playbook['stats'] = summary

        first_time = not os.path.exists(self.output_path)

        if first_time:
            with open(self.output_path, 'w') as outfile:
                outfile.write('[\n\n]\n')

        with open(self.output_path, 'r+') as outfile:
            self._append_playbook(outfile, first_time)

    def _append_playbook(self, outfile, first_time):
        file_len = outfile.seek(0, os.SEEK_END)
        # Remove three bytes to eat the trailing newline written by the
        # json.dump. This puts the ',' on the end of lines.
        outfile.seek(file_len - 3)
        if not first_time:
            outfile.write(',\n')
        json.dump(self.playbook, outfile,
                  indent=4, sort_keys=True, separators=(',', ': '))
        outfile.write('\n]\n')

    v2_runner_on_unreachable = v2_runner_on_ok


# Using 'ansible.builtin.command' instead of 'command' bypasses our
# custom plugins, so rewrite any uses of ansible.builtin.X to just X.
# This workaround is temporary until we remove our custom plugins.

# This happens here because Ansible will load the zuul_json plugin for
# any invocation where we care about restricting access, and this is
# the earliest in the Ansible startup procedure we can access.

# Monkepatch some PluginLoader methods to rewrite the modules it is
# loading.
orig_get = PluginLoader.get
orig_find_plugin = PluginLoader.find_plugin


def mp_get(self, name, *args, **kwargs):
    if (name.startswith('ansible.builtin.') or
        name.startswith('ansible.legacy.')):
        name = name.rsplit('.', 1)[-1]
    ret = orig_get(self, name, *args, **kwargs)
    return ret


def mp_find_plugin(self, name, *args, **kwargs):
    if (name.startswith('ansible.builtin.') or
        name.startswith('ansible.legacy.')):
        name = name.rsplit('.', 1)[-1]
    ret = orig_find_plugin(self, name, *args, **kwargs)
    return ret


PluginLoader.get = mp_get
PluginLoader.find_plugin = mp_find_plugin



#!/usr/bin/env python3

import logging
from collections import UserDict

from zuul.driver.github.githubconnection import GithubConnection
from zuul.driver.github import GithubDriver
from zuul.model import Change
from zuul.zk.change_cache import ChangeKey

# This is a template with boilerplate code for debugging github issues

# TODO: for real use override the following variables
server = 'github.com'
api_token = 'xxxx'
appid = 2
appkey = '/opt/project/appkey'

org = 'example'
repo = 'sandbox'
pull_nr = 8


class DummyChangeCache(UserDict):

    def updateChangeWithRetry(self, key, change, update_func, retry_count=5):
        update_func(change)
        self[key] = change
        return change


def configure_logging(context):
    stream_handler = logging.StreamHandler()
    logger = logging.getLogger(context)
    logger.addHandler(stream_handler)
    logger.setLevel(logging.DEBUG)


# uncomment for more logging
# configure_logging('urllib3')
# configure_logging('github3')
# configure_logging('cachecontrol')


# This is all that's needed for getting a usable github connection
def create_connection(server, api_token):
    driver = GithubDriver()
    connection_config = {
        'server': server,
        'api_token': api_token,
    }
    conn = GithubConnection(driver, 'github', connection_config)
    conn._github_client_manager.initialize()
    conn._change_cache = DummyChangeCache()
    return conn


def create_connection_app(server, appid, appkey):
    driver = GithubDriver()
    connection_config = {
        'server': server,
        'app_id': appid,
        'app_key': appkey,
    }
    conn = GithubConnection(driver, 'github', connection_config)
    conn._github_client_manager.initialize()
    conn._change_cache = DummyChangeCache()
    return conn


def get_change(connection: GithubConnection,
               org: str,
               repo: str,
               pull: int) -> Change:
    project_name = f"{org}/{repo}"
    github = connection.getGithubClient(project_name)
    pr = github.pull_request(org, repo, pull)
    sha = pr.head.sha
    change_key = ChangeKey('github', project_name, 'PullRequest', pull, sha)
    return conn._getChange(change_key, refresh=True)


# create github connection with api token
conn = create_connection(server, api_token)

# create github connection with app key
# conn = create_connection_app(server, appid, appkey)


# Now we can do anything we want with the connection, e.g. check canMerge for
# a pull request.
change = get_change(conn, org, repo, pull_nr)

print(conn.canMerge(change, {'cc/gate2'}))


# Or just use the github object.
# github = conn.getGithubClient()
#
# repository = github.repository(org, repo)
# print(repository.as_dict())



# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import configparser
import uuid

from zuul import model
from zuul.lib import tracing
import zuul.nodepool

from tests.base import (
    BaseTestCase,
    FakeNodepool,
    iterate_timeout,
    ZOOKEEPER_SESSION_TIMEOUT,
)
from zuul.zk import ZooKeeperClient
from zuul.zk.nodepool import ZooKeeperNodepool


class NodepoolWithCallback(zuul.nodepool.Nodepool):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.provisioned_requests = []

    def _handleNodeRequestEvent(self, request, event):
        super()._handleNodeRequestEvent(request, event)
        self.provisioned_requests.append(request)


class Dummy(object):
    def __init__(self, **kw):
        for k, v in kw.items():
            setattr(self, k, v)


class TestNodepoolBase(BaseTestCase):
    # Tests the Nodepool interface class using a fake nodepool and
    # scheduler.

    def setUp(self):
        super().setUp()

        config = configparser.ConfigParser()
        self.tracing = tracing.Tracing(config)
        self.statsd = None
        self.setupZK()

        self.zk_client = ZooKeeperClient(
            self.zk_chroot_fixture.zk_hosts,
            tls_cert=self.zk_chroot_fixture.zookeeper_cert,
            tls_key=self.zk_chroot_fixture.zookeeper_key,
            tls_ca=self.zk_chroot_fixture.zookeeper_ca,
            timeout=ZOOKEEPER_SESSION_TIMEOUT,
        )
        self.zk_nodepool = ZooKeeperNodepool(self.zk_client)
        self.addCleanup(self.zk_client.disconnect)
        self.zk_client.connect()
        self.hostname = 'nodepool-test-hostname'

        self.nodepool = NodepoolWithCallback(
            self.zk_client, self.hostname, self.statsd, scheduler=True)

        self.fake_nodepool = FakeNodepool(self.zk_chroot_fixture)
        self.addCleanup(self.fake_nodepool.stop)


class TestNodepool(TestNodepoolBase):
    def test_node_request(self):
        # Test a simple node request

        nodeset = model.NodeSet()
        nodeset.addNode(model.Node(['controller', 'foo'], 'ubuntu-xenial'))
        nodeset.addNode(model.Node(['compute'], 'ubuntu-xenial'))
        job = model.Job('testjob')
        job.uuid = uuid.uuid4().hex
        job.nodeset = nodeset
        request = self.nodepool.requestNodes(
            "test-uuid", job, "tenant", "pipeline", "provider", 0, 0)
        for x in iterate_timeout(30, 'requests are complete'):
            if len(self.nodepool.provisioned_requests) == 1:
                break
        request = self.nodepool.zk_nodepool.getNodeRequest(request.id)
        # We have to look up the request from ZK directly to check the
        # state.
        zk_request = self.zk_nodepool.getNodeRequest(request.id)
        self.assertEqual(zk_request.state, 'fulfilled')

        # Accept the nodes
        new_nodeset_info = self.nodepool.getNodesetInfo(request, nodeset)
        self.assertIsNotNone(new_nodeset_info)
        # acceptNodes will be called on the executor, but only if the
        # noderequest was accepted before.
        executor_nodeset = nodeset.copy()
        self.nodepool.acceptNodes(request, executor_nodeset)

        for node in executor_nodeset.getNodes():
            self.assertIsNotNone(node.lock)
            self.assertEqual(node.state, 'ready')

        # Mark the nodes in use
        self.nodepool.useNodeSet(
            executor_nodeset, tenant_name=None, project_name=None)
        for node in executor_nodeset.getNodes():
            self.assertEqual(node.state, 'in-use')

        # Return the nodes
        self.nodepool.returnNodeSet(
            executor_nodeset, build=None, tenant_name=None, project_name=None,
            duration=0)
        for node in executor_nodeset.getNodes():
            self.assertIsNone(node.lock)
            self.assertEqual(node.state, 'used')

    def test_node_request_canceled(self):
        # Test that node requests can be canceled

        nodeset = model.NodeSet()
        nodeset.addNode(model.Node(['controller'], 'ubuntu-xenial'))
        nodeset.addNode(model.Node(['compute'], 'ubuntu-xenial'))
        job = model.Job('testjob')
        job.uuid = uuid.uuid4().hex
        job.nodeset = nodeset
        self.fake_nodepool.pause()
        request = self.nodepool.requestNodes(
            "test-uuid", job, "tenant", "pipeline", "provider", 0, 0)
        for x in iterate_timeout(30, 'request created'):
            if len(self.nodepool.zk_nodepool._node_request_cache):
                break

        self.assertEqual(len(self.nodepool.provisioned_requests), 0)
        self.nodepool.cancelRequest(request)

        for x in iterate_timeout(30, 'request deleted'):
            if len(self.nodepool.provisioned_requests):
                break

        self.assertEqual(len(self.nodepool.provisioned_requests), 1)
        self.assertEqual(self.nodepool.provisioned_requests[0].state,
                         'requested')

    def test_node_request_priority(self):
        # Test that requests are satisfied in priority order

        nodeset = model.NodeSet()
        nodeset.addNode(model.Node(['controller', 'foo'], 'ubuntu-xenial'))
        nodeset.addNode(model.Node(['compute'], 'ubuntu-xenial'))
        job = model.Job('testjob')
        job.uuid = uuid.uuid4().hex
        job.nodeset = nodeset
        self.fake_nodepool.pause()
        request1 = self.nodepool.requestNodes(
            "test-uuid", job, "tenant", "pipeline", "provider", 0, 1)
        request2 = self.nodepool.requestNodes(
            "test-uuid", job, "tenant", "pipeline", "provider", 0, 0)
        self.fake_nodepool.unpause()
        for x in iterate_timeout(30, 'requests are complete'):
            if len(self.nodepool.provisioned_requests) == 2:
                break
        request1 = self.nodepool.zk_nodepool.getNodeRequest(request1.id)
        request2 = self.nodepool.zk_nodepool.getNodeRequest(request2.id)
        self.assertEqual(request1.state, 'fulfilled')
        self.assertEqual(request2.state, 'fulfilled')
        self.assertTrue(request2.state_time < request1.state_time)



# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import base64
import os

from zuul.lib import yamlutil as yaml

from tests.base import AnsibleZuulTestCase
from tests.base import ZuulTestCase


class TestInventoryBase(ZuulTestCase):

    config_file = 'zuul-gerrit-github.conf'
    tenant_config_file = 'config/inventory/main.yaml'
    use_gerrit = True

    def setUp(self, python_path=None, shell_type=None):
        super(TestInventoryBase, self).setUp()
        if python_path:
            self.fake_nodepool.python_path = python_path
        if shell_type:
            self.fake_nodepool.shell_type = shell_type
        self.executor_server.hold_jobs_in_build = True
        self.hold_jobs_in_queue = True

        if self.use_gerrit:
            A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
            self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        else:
            A = self.fake_github.openFakePullRequest(
                'org/project3', 'master', 'A')
            self.fake_github.emitEvent(A.getPullRequestOpenedEvent())
        self.change_A = A

        self.waitUntilSettled()

    def tearDown(self):
        self.cancelExecutorJobs()
        self.waitUntilSettled()
        super(TestInventoryBase, self).tearDown()

    def _get_build_inventory(self, name):
        self.runJob(name)

        build = self.getBuildByName(name)
        inv_path = os.path.join(build.jobdir.root, 'ansible', 'inventory.yaml')
        with open(inv_path, 'r') as f:
            inventory = yaml.safe_load(f)
        return inventory

    def _get_setup_inventory(self, name):
        self.runJob(name)

        build = self.getBuildByName(name)
        setup_inv_path = build.jobdir.setup_playbook.inventory
        with open(setup_inv_path, 'r') as f:
            inventory = yaml.ansible_unsafe_load(f)
        return inventory

    def runJob(self, name):
        self.hold_jobs_in_queue = False
        self.executor_api.release(f'^{name}$')
        self.waitUntilSettled()

    def cancelExecutorJobs(self):
        if self.use_gerrit:
            self.fake_gerrit.addEvent(
                self.change_A.getChangeAbandonedEvent())
        else:
            self.fake_github.emitEvent(
                self.change_A.getPullRequestClosedEvent())


class TestInventoryGithub(TestInventoryBase):

    use_gerrit = False

    def test_single_inventory(self):

        inventory = self._get_build_inventory('single-inventory')

        all_nodes = ('ubuntu-xenial',)
        self.assertIn('all', inventory)
        self.assertIn('hosts', inventory['all'])
        self.assertIn('vars', inventory['all'])
        for node_name in all_nodes:
            self.assertIn(node_name, inventory['all']['hosts'])
            node_vars = inventory['all']['hosts'][node_name]
            self.assertEqual(
                '/usr/bin/python3', node_vars['ansible_python_interpreter'])
        self.assertIn('zuul', inventory['all']['vars'])
        self.assertIn('attempts', inventory['all']['vars']['zuul'])
        self.assertEqual(1, inventory['all']['vars']['zuul']['attempts'])
        self.assertIn('max_attempts', inventory['all']['vars']['zuul'])
        self.assertEqual(3, inventory['all']['vars']['zuul']['max_attempts'])
        z_vars = inventory['all']['vars']['zuul']
        self.assertIn('executor', z_vars)
        self.assertIn('src_root', z_vars['executor'])
        self.assertIn('ansible_version', z_vars)
        self.assertIn('job', z_vars)
        self.assertIn('event_id', z_vars)
        self.assertEqual(z_vars['job'], 'single-inventory')
        self.assertEqual(z_vars['message'], 'QQ==')
        self.assertEqual(z_vars['change_url'],
                         'https://github.com/org/project3/pull/1')

        self.executor_server.release()
        self.waitUntilSettled()


class TestInventoryPythonPath(TestInventoryBase):

    def setUp(self):
        super(TestInventoryPythonPath, self).setUp(python_path='fake-python')

    def test_single_inventory(self):
        inventory = self._get_build_inventory('single-inventory')

        all_nodes = ('ubuntu-xenial',)
        self.assertIn('all', inventory)
        self.assertIn('hosts', inventory['all'])
        self.assertIn('vars', inventory['all'])
        for node_name in all_nodes:
            self.assertIn(node_name, inventory['all']['hosts'])
            node_vars = inventory['all']['hosts'][node_name]
            self.assertEqual(
                'fake-python', node_vars['ansible_python_interpreter'])

        self.assertIn('zuul', inventory['all']['vars'])
        z_vars = inventory['all']['vars']['zuul']
        self.assertIn('executor', z_vars)
        self.assertIn('src_root', z_vars['executor'])
        self.assertIn('ansible_version', z_vars)
        self.assertIn('job', z_vars)
        self.assertEqual(z_vars['job'], 'single-inventory')
        self.assertEqual(z_vars['message'], 'QQ==')

        self.executor_server.release()
        self.waitUntilSettled()


class TestInventoryShellType(TestInventoryBase):

    def setUp(self):
        super(TestInventoryShellType, self).setUp(shell_type='cmd')

    def test_single_inventory(self):
        inventory = self._get_build_inventory('single-inventory')

        all_nodes = ('ubuntu-xenial',)
        self.assertIn('all', inventory)
        self.assertIn('hosts', inventory['all'])
        self.assertIn('vars', inventory['all'])
        for node_name in all_nodes:
            self.assertIn(node_name, inventory['all']['hosts'])
            node_vars = inventory['all']['hosts'][node_name]
            self.assertEqual(
                'cmd', node_vars['ansible_shell_type'])

        self.assertIn('zuul', inventory['all']['vars'])
        z_vars = inventory['all']['vars']['zuul']
        self.assertIn('executor', z_vars)
        self.assertIn('src_root', z_vars['executor'])
        self.assertIn('ansible_version', z_vars)
        self.assertIn('job', z_vars)
        self.assertEqual(z_vars['job'], 'single-inventory')
        self.assertEqual(z_vars['message'], 'QQ==')

        self.executor_server.release()
        self.waitUntilSettled()


class InventoryAutoPythonMixin:
    ansible_version = 'X'

    def test_auto_python_ansible_inventory(self):
        inventory = self._get_build_inventory(
            f'ansible-version{self.ansible_version}-inventory')

        all_nodes = ('ubuntu-xenial',)
        self.assertIn('all', inventory)
        self.assertIn('hosts', inventory['all'])
        self.assertIn('vars', inventory['all'])
        for node_name in all_nodes:
            self.assertIn(node_name, inventory['all']['hosts'])
            node_vars = inventory['all']['hosts'][node_name]
            self.assertEqual(
                '/usr/bin/python3', node_vars['ansible_python_interpreter'])

        self.assertIn('zuul', inventory['all']['vars'])
        z_vars = inventory['all']['vars']['zuul']
        self.assertIn('executor', z_vars)
        self.assertIn('src_root', z_vars['executor'])
        self.assertIn('job', z_vars)
        self.assertEqual(z_vars['job'],
                         f'ansible-version{self.ansible_version}-inventory')
        self.assertEqual(z_vars['message'], 'QQ==')

        self.executor_server.release()
        self.waitUntilSettled()


class TestInventoryAutoPythonAnsible9(TestInventoryBase,
                                      InventoryAutoPythonMixin):
    ansible_version = '9'


class TestInventoryAutoPythonAnsible11(TestInventoryBase,
                                       InventoryAutoPythonMixin):
    ansible_version = '11'


class TestInventory(TestInventoryBase):

    def test_single_inventory(self):

        inventory = self._get_build_inventory('single-inventory')

        all_nodes = ('ubuntu-xenial',)
        self.assertIn('all', inventory)
        self.assertIn('hosts', inventory['all'])
        self.assertIn('vars', inventory['all'])
        for node_name in all_nodes:
            self.assertIn(node_name, inventory['all']['hosts'])
            node_vars = inventory['all']['hosts'][node_name]
            self.assertEqual(
                '/usr/bin/python3', node_vars['ansible_python_interpreter'])
            self.assertNotIn(
                'ansible_shell_type', node_vars)
        self.assertIn('zuul', inventory['all']['vars'])
        self.assertIn('attempts', inventory['all']['vars']['zuul'])
        self.assertEqual(1, inventory['all']['vars']['zuul']['attempts'])
        self.assertIn('max_attempts', inventory['all']['vars']['zuul'])
        self.assertEqual(3, inventory['all']['vars']['zuul']['max_attempts'])
        z_vars = inventory['all']['vars']['zuul']
        self.assertIn('executor', z_vars)
        self.assertIn('src_root', z_vars['executor'])
        self.assertIn('job', z_vars)
        self.assertEqual(z_vars['job'], 'single-inventory')
        self.assertEqual(z_vars['message'], 'QQ==')

        self.executor_server.release()
        self.waitUntilSettled()

    def test_single_inventory_list(self):

        inventory = self._get_build_inventory('single-inventory-list')

        all_nodes = ('compute', 'controller')
        self.assertIn('all', inventory)
        self.assertIn('hosts', inventory['all'])
        self.assertIn('vars', inventory['all'])
        for node_name in all_nodes:
            self.assertIn(node_name, inventory['all']['hosts'])
        self.assertIn('zuul', inventory['all']['vars'])
        z_vars = inventory['all']['vars']['zuul']
        self.assertIn('executor', z_vars)
        self.assertIn('src_root', z_vars['executor'])
        self.assertIn('job', z_vars)
        self.assertEqual(z_vars['job'], 'single-inventory-list')

        self.executor_server.release()
        self.waitUntilSettled()

    def test_executor_only_inventory(self):
        inventory = self._get_build_inventory('executor-only-inventory')

        self.assertIn('all', inventory)
        self.assertIn('hosts', inventory['all'])
        self.assertIn('vars', inventory['all'])

        # Should be blank; i.e. rely on the implicit localhost
        self.assertEqual(0, len(inventory['all']['hosts']))

        self.assertIn('zuul', inventory['all']['vars'])
        z_vars = inventory['all']['vars']['zuul']
        self.assertIn('executor', z_vars)
        self.assertIn('src_root', z_vars['executor'])
        self.assertIn('job', z_vars)
        self.assertEqual(z_vars['job'], 'executor-only-inventory')
        self.assertEqual(z_vars['message'], 'QQ==')

        self.executor_server.release()
        self.waitUntilSettled()

    def test_group_inventory(self):

        inventory = self._get_build_inventory('group-inventory')

        all_nodes = ('controller', 'compute1', 'compute2')
        self.assertIn('all', inventory)
        self.assertIn('children', inventory['all'])
        self.assertIn('hosts', inventory['all'])
        self.assertIn('vars', inventory['all'])
        for group_name in ('ceph-osd', 'ceph-monitor'):
            self.assertIn(group_name, inventory['all']['children'])
        for node_name in all_nodes:
            self.assertIn(node_name, inventory['all']['hosts'])
            self.assertIn(node_name,
                          inventory['all']['children']
                          ['ceph-monitor']['hosts'])
        self.assertEqual(
            '/usr/bin/python3',
            inventory['all']['hosts']['controller']
            ['ansible_python_interpreter'])
        self.assertEqual(
            'ceph',
            inventory['all']['hosts']['controller']
            ['ceph_var'])
        self.assertEqual(
            '/usr/bin/python3',
            inventory['all']['hosts']['compute1']
            ['ansible_python_interpreter'])
        self.assertNotIn(
            'ceph_var',
            inventory['all']['hosts']['compute1'])
        self.assertIn('zuul', inventory['all']['vars'])
        z_vars = inventory['all']['vars']['zuul']
        self.assertIn('executor', z_vars)
        self.assertIn('src_root', z_vars['executor'])
        self.assertIn('job', z_vars)
        self.assertEqual(z_vars['job'], 'group-inventory')

        self.executor_server.release()
        self.waitUntilSettled()

    def test_hostvars_inventory(self):

        inventory = self._get_build_inventory('hostvars-inventory')

        all_nodes = ('default', 'fakeuser')
        self.assertIn('all', inventory)
        self.assertIn('hosts', inventory['all'])
        self.assertIn('vars', inventory['all'])
        for node_name in all_nodes:
            self.assertIn(node_name, inventory['all']['hosts'])
            # check if the nodes use the correct username
            if node_name == 'fakeuser':
                username = 'fakeuser'
            else:
                username = 'zuul'
            self.assertEqual(
                inventory['all']['hosts'][node_name]['ansible_user'], username)

            # check if the nodes use the correct or no ansible_connection
            if node_name == 'windows':
                self.assertEqual(
                    inventory['all']['hosts'][node_name]['ansible_connection'],
                    'winrm')
            else:
                self.assertEqual(
                    'local',
                    inventory['all']['hosts'][node_name]['ansible_connection'])

            self.assertEqual(
                '/usr/bin/python3',
                inventory['all']['hosts'][node_name]
                ['ansible_python_interpreter'])
            self.assertEqual(
                'all',
                inventory['all']['hosts'][node_name]
                ['all_var'])
        self.assertNotIn(
            'ansible_python_interpreter',
            inventory['all']['vars'])

        self.executor_server.release()
        self.waitUntilSettled()

    def test_setup_inventory(self):

        setup_inventory = self._get_setup_inventory('hostvars-inventory')
        inventory = self._get_build_inventory('hostvars-inventory')

        self.assertIn('all', inventory)
        self.assertIn('hosts', inventory['all'])

        self.assertIn('default', setup_inventory['all']['hosts'])
        self.assertIn('fakeuser', setup_inventory['all']['hosts'])
        self.assertIn('windows', setup_inventory['all']['hosts'])
        self.assertNotIn('network', setup_inventory['all']['hosts'])
        self.assertIn('default', inventory['all']['hosts'])
        self.assertIn('fakeuser', inventory['all']['hosts'])
        self.assertIn('windows', inventory['all']['hosts'])
        self.assertIn('network', inventory['all']['hosts'])

        self.executor_server.release()
        self.waitUntilSettled()


class TestAnsibleInventory(AnsibleZuulTestCase):

    config_file = 'zuul-gerrit-github.conf'
    tenant_config_file = 'config/inventory/main.yaml'

    def _get_file(self, build, path):
        p = os.path.join(build.jobdir.root, path)
        with open(p) as f:
            return f.read()

    def _jinja2_message(self, expected_message):

        # This test runs a bit long and needs extra time.
        self.wait_timeout = 120
        # Keep the jobdir around to check inventory
        self.executor_server.keep_jobdir = True
        # Output extra ansible info so we might see errors.
        self.executor_server.verbose = True
        A = self.fake_gerrit.addFakeChange(
            'org/project2', 'master', expected_message,
            files={'jinja.txt': 'foo'})
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertHistory([
            dict(name='jinja2-message', result='SUCCESS', changes='1,1')])

        build = self.history[0]
        inv_path = os.path.join(build.jobdir.root, 'ansible', 'inventory.yaml')
        with open(inv_path, 'r') as f:
            inventory = yaml.safe_load(f)

        zv_path = os.path.join(build.jobdir.root, 'ansible', 'zuul_vars.yaml')
        with open(zv_path, 'r') as f:
            zv = yaml.ansible_unsafe_load(f)

        # TODO(corvus): zuul vars aren't really stored here anymore;
        # rework these tests to examine them separately.
        inventory['all']['vars'] = {'zuul': zv['zuul']}

        # The deprecated base64 version
        decoded_message = base64.b64decode(
            inventory['all']['vars']['zuul']['message']).decode('utf-8')
        self.assertEqual(decoded_message, expected_message)
        obtained_message = self._get_file(self.history[0],
                                          'work/logs/commit-message.txt')
        self.assertEqual(obtained_message, expected_message)

        # The new !unsafe version
        decoded_message = inventory['all']['vars']['zuul']['change_message']
        self.assertEqual(decoded_message, expected_message)
        obtained_message = self._get_file(self.history[0],
                                          'work/logs/change-message.txt')
        self.assertEqual(obtained_message, expected_message)

    def test_jinja2_message_brackets(self):
        self._jinja2_message("This message has {{ ansible_host }} in it ")

    def test_jinja2_message_raw(self):
        self._jinja2_message("This message has {% raw %} in {% endraw %} it ")

    def test_network_inventory(self):
        # Network appliances can't run the freeze or setup playbooks,
        # so they won't have any job variables available.  But they
        # should still have nodepool hostvars.  Run a playbook that
        # verifies that.
        A = self.fake_gerrit.addFakeChange(
            'org/project2', 'master', 'A',
            files={'network.txt': 'foo'})
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertHistory([
            dict(name='network', result='SUCCESS', changes='1,1')])


class TestWindowsInventory(TestInventoryBase):
    config_file = 'zuul-winrm.conf'

    def test_windows_inventory(self):

        inventory = self._get_build_inventory('hostvars-inventory')
        windows_host = inventory['all']['hosts']['windows']
        self.assertEqual(windows_host['ansible_connection'], 'winrm')
        self.assertEqual(
            windows_host['ansible_winrm_operation_timeout_sec'],
            '120')
        self.assertEqual(
            windows_host['ansible_winrm_read_timeout_sec'],
            '180')

        self.executor_server.release()
        self.waitUntilSettled()



# Copyright 2020 BMW Group
# Copyright 2022, 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
from collections.abc import MutableMapping
from contextlib import suppress
from functools import total_ordering
import logging
import time
import zlib

from kazoo.exceptions import NoNodeError

from zuul.zk import sharding, ZooKeeperBase, ZooKeeperSimpleBase
from zuul.provider import BaseProvider


@total_ordering
class LayoutState:
    """Representation of a tenant's layout state.

    The layout state holds information about a certain version of a
    tenant's layout. It is used to coordinate reconfigurations across
    multiple schedulers by comparing a local tenant layout state
    against the current version in Zookeeper. In case it detects that
    a local layout state is outdated, this scheduler is not allowed to
    process this tenant (events, pipelines, ...) until the layout is
    updated.

    The important information of the layout state is the logical
    timestamp (ltime) that is used to detect if the layout on a
    scheduler needs to be updated. The ltime is the last modified
    transaction ID (mzxid) of the corresponding Znode in Zookeeper.

    The hostname of the scheduler creating the new layout state and the
    timestamp of the last reconfiguration are only informational and
    may aid in debugging.
    """

    def __init__(self, tenant_name, hostname, last_reconfigured, uuid,
                 branch_cache_min_ltimes, last_reconfigure_event_ltime,
                 ltime=-1):
        self.uuid = uuid
        self.ltime = ltime
        self.tenant_name = tenant_name
        self.hostname = hostname
        self.last_reconfigured = last_reconfigured
        self.last_reconfigure_event_ltime =\
            last_reconfigure_event_ltime
        self.branch_cache_min_ltimes = branch_cache_min_ltimes

    def toDict(self):
        return {
            "tenant_name": self.tenant_name,
            "hostname": self.hostname,
            "last_reconfigured": self.last_reconfigured,
            "last_reconfigure_event_ltime":
            self.last_reconfigure_event_ltime,
            "uuid": self.uuid,
            "branch_cache_min_ltimes": self.branch_cache_min_ltimes,
        }

    @classmethod
    def fromDict(cls, data):
        return cls(
            data["tenant_name"],
            data["hostname"],
            data["last_reconfigured"],
            data.get("uuid"),
            data.get("branch_cache_min_ltimes"),
            data.get("last_reconfigure_event_ltime", -1),
            data.get("ltime", -1),
        )

    def __eq__(self, other):
        if not isinstance(other, LayoutState):
            return False
        return self.uuid == other.uuid

    def __gt__(self, other):
        if not isinstance(other, LayoutState):
            return False
        return self.ltime > other.ltime

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} {self.tenant_name}: "
            f"ltime={self.ltime}, "
            f"hostname={self.hostname}, "
            f"last_reconfigured={self.last_reconfigured}>"
        )


class LayoutStateStore(ZooKeeperBase, MutableMapping):
    log = logging.getLogger("zuul.LayoutStore")

    layout_root = "/zuul/layout"
    layout_data_root = "/zuul/layout-data"

    def __init__(self, client, callback):
        super().__init__(client)
        self._watched_tenants = set()
        self._callback = callback
        self.kazoo_client.ensure_path(self.layout_root)
        self.kazoo_client.ChildrenWatch(self.layout_root, self._layoutCallback)

    def _layoutCallback(self, tenant_list, event=None):
        new_tenants = set(tenant_list) - self._watched_tenants
        for tenant_name in new_tenants:
            self.kazoo_client.DataWatch(f"{self.layout_root}/{tenant_name}",
                                        self._callbackWrapper)

    def _callbackWrapper(self, data, stat, event):
        self._callback()

    def __getitem__(self, tenant_name):
        try:
            data, zstat = self.kazoo_client.get(
                f"{self.layout_root}/{tenant_name}")
        except NoNodeError:
            raise KeyError(tenant_name)

        if not data:
            raise KeyError(tenant_name)

        return LayoutState.fromDict({
            "ltime": zstat.last_modified_transaction_id,
            **json.loads(data)
        })

    def __setitem__(self, tenant_name, state):
        path = f"{self.layout_root}/{tenant_name}"
        data = json.dumps(state.toDict(), sort_keys=True).encode("utf-8")
        if self.kazoo_client.exists(path):
            zstat = self.kazoo_client.set(path, data)
        else:
            _, zstat = self.kazoo_client.create(path, data, include_data=True)
        # Set correct ltime of the layout in Zookeeper
        state.ltime = zstat.last_modified_transaction_id

    def __delitem__(self, tenant_name):
        try:
            self.kazoo_client.delete(f"{self.layout_root}/{tenant_name}")
        except NoNodeError:
            raise KeyError(tenant_name)

    def __iter__(self):
        try:
            tenant_names = self.kazoo_client.get_children(self.layout_root)
        except NoNodeError:
            return
        yield from tenant_names

    def __len__(self):
        zstat = self.kazoo_client.exists(self.layout_root)
        if zstat is None:
            return 0
        return zstat.children_count

    def getMinLtimes(self, layout_state):
        try:
            path = f"{self.layout_data_root}/{layout_state.uuid}"
            with sharding.BufferedShardReader(
                    self.kazoo_client, path) as stream:
                data = zlib.decompress(stream.read())
        except NoNodeError:
            return None

        try:
            return json.loads(data)['min_ltimes']
        except Exception:
            return None

    def setMinLtimes(self, layout_state, min_ltimes):
        data = dict(min_ltimes=min_ltimes)
        encoded_data = json.dumps(data).encode("utf-8")

        path = f"{self.layout_data_root}/{layout_state.uuid}"
        with sharding.BufferedShardWriter(
                self.kazoo_client, path) as stream:
            stream.write(zlib.compress(encoded_data))

    def cleanup(self, delay=300):
        self.log.debug("Starting layout data cleanup")
        known_layouts = set()
        for tenant in self.kazoo_client.get_children(
                self.layout_root):
            state = self.get(tenant)
            if state:
                known_layouts.add(state.uuid)

        for layout_uuid in self.kazoo_client.get_children(
                self.layout_data_root):
            if layout_uuid in known_layouts:
                continue

            path = f"{self.layout_data_root}/{layout_uuid}"
            zstat = self.kazoo_client.exists(path)
            if zstat is None:
                continue

            now = time.time()
            if now - zstat.created >= delay:
                self.log.debug("Deleting unused layout data for %s",
                               layout_uuid)
                with suppress(NoNodeError):
                    self.kazoo_client.delete(path, recursive=True)
        self.log.debug("Finished layout data cleanup")


class LayoutProvidersStore(ZooKeeperSimpleBase):
    log = logging.getLogger("zuul.LayoutProvidersStore")

    tenant_root = "/zuul/tenant"

    def __init__(self, client, connections, system_id):
        super().__init__(client)
        self.connections = connections
        self.system_id = system_id

    def get(self, context, zk_client, tenant_name):
        path = f"{self.tenant_root}/{tenant_name}/provider"
        try:
            repo_names = self.kazoo_client.get_children(path)
        except NoNodeError:
            repo_names = []
        for repo in repo_names:
            path = f"{self.tenant_root}/{tenant_name}/provider/{repo}"
            try:
                provider_names = self.kazoo_client.get_children(path)
            except NoNodeError:
                provider_names = []
            for provider_name in provider_names:
                provider_path = (f"{path}/{provider_name}/config")
                yield BaseProvider.fromZK(
                    context, provider_path, self.connections,
                    self.system_id, zk_client
                )

    def set(self, context, tenant_name, providers):
        self.clear(tenant_name)
        path = f"{self.tenant_root}/{tenant_name}/provider"
        self.kazoo_client.ensure_path(path)
        for provider in providers:
            provider.internalCreate(context)

    def clear(self, tenant_name):
        path = f"{self.tenant_root}/{tenant_name}/provider"
        self.kazoo_client.delete(path, recursive=True)



# Copyright 2018 BMW Carit GmbH
#
# Zuul is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Zuul is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

# This is not needed in python3 - but it is needed in python2 because there
# is a json module in ansible.plugins.callback and python2 gets confused.
# Easy local testing with ansible-playbook is handy when hacking on zuul_stream
# so just put in the __future__ statement.
from __future__ import absolute_import

DOCUMENTATION = '''
     callback: zuul_unreachable
     short_description: Zuul Unreachable
     description:
         - Zuul Unreachable
     type: aggregate
     extends_documentation_fragment:
       - default_callback
'''

import os

from ansible.plugins.callback import default


class CallbackModule(default.CallbackModule):

    CALLBACK_VERSION = 2.0
    # aggregate means we can be loaded and not be the stdout plugin
    CALLBACK_TYPE = 'aggregate'
    CALLBACK_NAME = 'zuul_unreachable'

    def __init__(self):
        super(CallbackModule, self).__init__()
        self.output_path = os.path.join(
            os.environ['ZUUL_JOBDIR'], '.ansible', 'nodes.unreachable')
        self.unreachable_hosts = set()

    def v2_runner_on_unreachable(self, result):
        host = result._host.get_name()
        if host not in self.unreachable_hosts:
            self.unreachable_hosts.add(host)
            with open(self.output_path, 'a') as f:
                f.write('%s\n' % host)



# Copyright 2017 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from zuul.driver import Driver, ConnectionInterface, TriggerInterface
from zuul.driver import SourceInterface, ReporterInterface
from zuul.driver.github import githubconnection
from zuul.driver.github import githubmodel
from zuul.driver.github import githubtrigger
from zuul.driver.github import githubsource
from zuul.driver.github import githubreporter


class GithubDriver(Driver, ConnectionInterface, TriggerInterface,
                   SourceInterface, ReporterInterface):
    name = 'github'

    def getConnection(self, name, config):
        return githubconnection.GithubConnection(self, name, config)

    def getTrigger(self, connection, config=None):
        return githubtrigger.GithubTrigger(self, connection, config)

    def getTriggerEventClass(self):
        return githubmodel.GithubTriggerEvent

    def getSource(self, connection):
        return githubsource.GithubSource(self, connection)

    def getReporter(self, connection, pipeline, config=None,
                    parse_context=None):
        return githubreporter.GithubReporter(
            self, connection, pipeline, config, parse_context)

    def getTriggerSchema(self):
        return githubtrigger.getSchema()

    def getReporterSchema(self):
        return githubreporter.getSchema()

    def getRequireSchema(self):
        return githubsource.getRequireSchema()

    def getRejectSchema(self):
        return githubsource.getRejectSchema()



# Copyright 2024-2025 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import base64
import contextlib
import ipaddress
import time
from unittest import mock
import urllib.parse

import fixtures
from moto import mock_aws
import boto3
import botocore.exceptions
from kazoo.exceptions import NoNodeError

from zuul import model
from zuul.driver.aws import AwsDriver
from zuul.driver.aws.awsmodel import AwsProviderNode
import zuul.driver.aws.awsendpoint

from tests.fake_aws import FakeAws, FakeAwsProviderEndpoint
from tests.base import (
    AnsibleZuulTestCase,
    ResponsesFixture,
    driver_config,
    iterate_timeout,
    okay_tracebacks,
    return_data,
    simple_layout,
)
from tests.unit.test_launcher import ImageMocksFixture
from tests.unit.test_cloud_driver import BaseCloudDriverTest


def _make_ipv6_subnets(cidr_block):
    network = ipaddress.IPv6Network(cidr_block)
    # AWS only supports /64 prefix length
    return [str(sn) for sn in network.subnets(new_prefix=64)]


class AwsBaseTest(BaseCloudDriverTest):
    config_file = 'zuul-connections-nodepool.conf'
    cloud_test_image_format = 'raw'
    cloud_test_provider_name = 'aws-us-east-1-main'
    mock_aws = mock_aws()
    debian_return_data = {
        'zuul': {
            'artifacts': [
                {
                    'name': 'raw image',
                    'url': 'http://example.com/image.raw',
                    'metadata': {
                        'type': 'zuul_image',
                        'image_name': 'debian-local',
                        'format': 'raw',
                        'sha256': ImageMocksFixture.raw_sha256,
                        'md5sum': ImageMocksFixture.raw_md5sum,
                    }
                },
            ]
        }
    }
    s3_debian_return_data = {
        'zuul': {
            'artifacts': [
                {
                    'name': 'raw image',
                    'url': 's3://zuul/image.raw',
                    'metadata': {
                        'type': 'zuul_image',
                        'image_name': 'debian-local',
                        'format': 'raw',
                        'sha256': ImageMocksFixture.raw_sha256,
                        'md5sum': ImageMocksFixture.raw_md5sum,
                    }
                },
            ]
        }
    }
    s3_region_debian_return_data = {
        'zuul': {
            'artifacts': [
                {
                    'name': 'raw image',
                    'url': 's3://zuulwest/image.raw',
                    'metadata': {
                        'type': 'zuul_image',
                        'image_name': 'debian-local',
                        'format': 'raw',
                        'sha256': ImageMocksFixture.raw_sha256,
                        'md5sum': ImageMocksFixture.raw_md5sum,
                    }
                },
            ]
        }
    }

    def setUp(self):
        self.initTestConfig()
        aws_id = 'AK000000000000000000'
        aws_key = '0123456789abcdef0123456789abcdef0123456789abcdef'
        self.useFixture(
            fixtures.EnvironmentVariable('AWS_ACCESS_KEY_ID', aws_id))
        self.useFixture(
            fixtures.EnvironmentVariable('AWS_SECRET_ACCESS_KEY', aws_key))
        self.patch(zuul.driver.aws.awsendpoint, 'CACHE_TTL', 1)

        self.fake_aws = FakeAws()
        self.mock_aws.start()
        # Must start responses after mock_aws
        responses_fixture = self.useFixture(ResponsesFixture())
        self.useFixture(ImageMocksFixture(responses_fixture))

        self.ec2 = boto3.resource('ec2', region_name='us-east-1')
        self.ec2_client = boto3.client('ec2', region_name='us-east-1')
        self.s3 = boto3.resource('s3', region_name='us-east-1')
        self.s3_client = boto3.client('s3', region_name='us-east-1')
        self.iam = boto3.resource('iam', region_name='us-east-1')
        self.s3.create_bucket(Bucket='zuul')
        location = {'LocationConstraint': 'us-west-1'}
        self.s3.create_bucket(Bucket="zuulwest",
                              CreateBucketConfiguration=location)

        # A list of args to method calls for validation
        self.run_instances_calls = []
        self.run_instances_exception = None
        self.create_fleet_calls = []
        self.create_fleet_results = []
        self.create_fleet_exception = None
        self.allocate_hosts_exception = None
        self.register_image_calls = []
        self.copy_image_calls = []

        # TEST-NET-3
        self.subnet_ids = []
        ipv6 = False
        if ipv6:
            # This is currently unused, but if moto gains IPv6 support
            # on instance creation, this may be useful.
            self.vpc = self.ec2_client.create_vpc(
                CidrBlock='203.0.113.0/24',
                AmazonProvidedIpv6CidrBlock=True)
            ipv6_cidr = self.vpc['Vpc'][
                'Ipv6CidrBlockAssociationSet'][0]['Ipv6CidrBlock']
            ipv6_subnets = _make_ipv6_subnets(ipv6_cidr)

            subnet1 = self.ec2_client.create_subnet(
                AvailabilityZone='us-east-1a',
                CidrBlock='203.0.113.64/26',
                Ipv6CidrBlock=ipv6_subnets[0],
                VpcId=self.vpc['Vpc']['VpcId'])
            self.subnet_ids.append(subnet1['Subnet']['SubnetId'])
            subnet2 = self.ec2_client.create_subnet(
                AvailabilityZone='us-east-1b',
                CidrBlock='203.0.113.128/26',
                Ipv6CidrBlock=ipv6_subnets[1],
                VpcId=self.vpc['Vpc']['VpcId'])
            self.subnet_ids.append(subnet2['Subnet']['SubnetId'])
        else:
            self.vpc = self.ec2_client.create_vpc(CidrBlock='203.0.113.0/24')
            subnet1 = self.ec2_client.create_subnet(
                AvailabilityZone='us-east-1a',
                CidrBlock='203.0.113.64/26',
                VpcId=self.vpc['Vpc']['VpcId'])
            self.subnet_ids.append(subnet1['Subnet']['SubnetId'])
            subnet2 = self.ec2_client.create_subnet(
                AvailabilityZone='us-east-1b',
                CidrBlock='203.0.113.128/26',
                VpcId=self.vpc['Vpc']['VpcId'])
            self.subnet_ids.append(subnet2['Subnet']['SubnetId'])

        self.security_group = self.ec2_client.create_security_group(
            GroupName='zuul-nodes', VpcId=self.vpc['Vpc']['VpcId'],
            Description='Zuul Nodes')
        self.security_group_id = self.security_group['GroupId']
        self.profile = self.iam.create_instance_profile(
            InstanceProfileName='not-a-real-profile')

        self.patch(AwsDriver, '_endpoint_class', FakeAwsProviderEndpoint)
        self.patch(FakeAwsProviderEndpoint,
                   '_FakeAwsProviderEndpoint__testcase', self)

        default_ec2_quotas = {
            'L-1216C47A': 100,
            'L-43DA4232': 100,
            'L-34B43A08': 100,
        }
        default_ebs_quotas = {
            'L-D18FCD1D': 100.0,
            'L-7A658B76': 100.0,
        }
        ec2_quotas = self.test_config.driver.aws.get(
            'ec2_quotas', default_ec2_quotas)
        ebs_quotas = self.test_config.driver.aws.get(
            'ebs_quotas', default_ebs_quotas)
        self.patch(FakeAwsProviderEndpoint,
                   '_FakeAwsProviderEndpoint__ec2_quotas', ec2_quotas)
        self.patch(FakeAwsProviderEndpoint,
                   '_FakeAwsProviderEndpoint__ebs_quotas', ebs_quotas)
        self.lateSetUp()
        super().setUp()

    def lateSetUp(self):
        pass

    def shutdown(self):
        super().shutdown()
        self.mock_aws.stop()


class TestAwsDriver(AwsBaseTest):

    def _assertProviderNodeAttributes(self, pnode):
        super()._assertProviderNodeAttributes(pnode)
        if checks := self.test_config.driver.aws.get('node_checks'):
            checks(self, pnode)

    def check_node_attrs(self, pnode):
        self.assertEqual(
            1000,
            self.run_instances_calls[0]['BlockDeviceMappings'][0]['Ebs']
            ['Iops'])
        self.assertEqual(
            200,
            self.run_instances_calls[0]['BlockDeviceMappings'][0]['Ebs']
            ['Throughput'])
        self.assertEqual(
            'disabled',
            self.run_instances_calls[0]['MaintenanceOptions']
            ['AutoRecovery'])
        for tag_spec in self.run_instances_calls[0]['TagSpecifications']:
            tags = {t["Key"]: t["Value"] for t in tag_spec['Tags']}
            self.assertEqual(
                "tenant-one",
                tags["tenant"])
            self.assertIn(
                "event_id",
                tags)

    @simple_layout('layouts/nodepool.yaml', enable_nodepool=True)
    @driver_config('aws', node_checks=check_node_attrs)
    def test_aws_node_lifecycle(self):
        self._test_node_lifecycle('debian-normal')

    def check_spot_node_attrs(self, pnode):
        # The basic test above sets few options; we set many more
        # options in the spot check (so that we don't have run a test
        # for every option).
        self.assertEqual(
            'spot',
            self.run_instances_calls[0]['InstanceMarketOptions']['MarketType'])
        self.assertEqual(
            'us-east-1b',
            self.run_instances_calls[0]['Placement']['AvailabilityZone'])
        self.assertEqual(
            ['testgroup'],
            self.run_instances_calls[0]['NetworkInterfaces'][0]['Groups'])
        self.assertEqual(
            1,
            self.run_instances_calls[0]['NetworkInterfaces'][0][
                'Ipv6AddressCount'])
        self.assertEqual(
            'required',
            self.run_instances_calls[0]['MetadataOptions']['HttpTokens'])
        self.assertEqual(
            self.run_instances_calls[0]['BlockDeviceMappings'][0]['Ebs']
            ['KmsKeyId'], 'alias/aws/ebs')
        self.assertEqual(
            self.run_instances_calls[0]['BlockDeviceMappings'][0]['Ebs']
            ['Encrypted'], True)
        self.assertTrue(pnode.node_properties['spot'])
        instance = self.ec2_client.describe_instance_attribute(
            InstanceId=pnode.aws_instance_id,
            Attribute='userData',
        )
        expected = base64.b64encode(b'testuserdata').decode('utf8')
        self.assertEqual(expected, instance['UserData']['Value'])

    @simple_layout('layouts/aws/spot.yaml', enable_nodepool=True,
                   replace=lambda test: {
                       'subnet_ids': test.subnet_ids,
                       'iam_profile_name': test.profile.name,
                   })
    @driver_config('aws', node_checks=check_spot_node_attrs)
    def test_aws_node_lifecycle_spot(self):
        self._test_node_lifecycle('debian-normal')

    def check_fleet_node_attrs(self, pnode):
        self.assertEqual(
            'price-capacity-optimized',
            self.create_fleet_calls[0]['OnDemandOptions'][
                'AllocationStrategy'])
        self.assertTrue(pnode.node_properties['fleet'])
        instance = self.ec2_client.describe_instance_attribute(
            InstanceId=pnode.aws_instance_id,
            Attribute='userData',
        )
        expected = base64.b64encode(b'testuserdata').decode('utf8')
        self.assertEqual(expected, instance['UserData']['Value'])

    @simple_layout('layouts/aws/fleet.yaml', enable_nodepool=True)
    @okay_tracebacks('MissingParameter')
    @driver_config('aws', node_checks=check_fleet_node_attrs)
    def test_aws_node_lifecycle_fleet(self):
        self.log.debug("Start cleanup worker")
        self.launcher.cleanup_worker.INTERVAL = 1
        self.launcher.cleanup_worker.start()
        self._test_node_lifecycle('debian-normal')
        self.waitUntilSettled()

        # Verify that we clean up unused launch templates.  Start by
        # checking that we have one from the current config.
        launch_tempaltes = self.ec2_client.\
            describe_launch_templates()['LaunchTemplates']
        self.assertEqual(len(launch_tempaltes), 1)

        # Switch to a config that has no fleet usage (spot is
        # arbitrary).
        self.commitConfigUpdate(
            'org/common-config', 'layouts/aws/spot.yaml',
            replace=lambda test: {
                'subnet_ids': test.subnet_ids,
                'iam_profile_name': test.profile.name,
            })

        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))
        self.waitUntilSettled()

        # Verify that there are no launch templates.
        for _ in iterate_timeout(30, 'template deletion'):
            launch_templates = self.ec2_client.\
                describe_launch_templates()['LaunchTemplates']
            if len(launch_templates) == 0:
                break

    @simple_layout('layouts/nodepool.yaml', enable_nodepool=True)
    @driver_config('aws', ec2_quotas={
        'L-1216C47A': 2,
    })
    def test_aws_quota(self):
        self._test_quota('debian-normal')

    @simple_layout('layouts/aws/resource-limits.yaml', enable_nodepool=True)
    def test_aws_resource_limits(self):
        self._test_quota('debian-normal')

    @simple_layout('layouts/aws/nodepool-image-snapshot.yaml',
                   enable_nodepool=True)
    @return_data(
        'build-debian-local-image',
        'refs/heads/master',
        AwsBaseTest.debian_return_data,
    )
    def test_aws_diskimage_snapshot(self):
        self._test_diskimage()

    @simple_layout('layouts/aws/nodepool-image-snapshot.yaml',
                   enable_nodepool=True)
    @return_data(
        'build-debian-local-image',
        'refs/heads/master',
        AwsBaseTest.debian_return_data,
    )
    def test_aws_diskimage_snapshot_resume(self):
        self.addImageBuildEvent(
            'tenant-one',
            'review.example.com/org/common-config',
            'master',
            ['debian-local'],
        )
        self.waitUntilSettled()

        system_id = self.launcher.system.system_id
        image_tags = [
            {'Key': 'zuul_system_id', 'Value': system_id},
            {'Key': 'zuul_upload_uuid', 'Value': 'upload-uuid'},
        ]
        task = self.fake_aws.import_snapshot(
            DiskContainer={
                'Format': 'ova',
                'UserBucket': {
                    'S3Bucket': 'zuul',
                    'S3Key': 'testfile',
                }
            },
            TagSpecifications=[{
                'ResourceType': 'import-snapshot-task',
                'Tags': image_tags,
            }])
        self.fake_aws.finish_import_snapshot(task)

        provider = self.launcher._getProvider(
            'tenant-one', 'aws-us-east-1-main')
        endpoint = provider.getEndpoint()
        image = list(provider.images.values())[0]
        self.assertEqual('debian-local', image.name)
        # create an IBA and an upload
        with self.createZKContext(None) as ctx:
            # This starts with an unknown state, then
            # createImageUploads will set it to ready.
            iba = model.ImageBuildArtifact.new(
                ctx,
                uuid='iba-uuid',
                name=image.name,
                canonical_name=image.canonical_name,
                project_canonical_name=image.project_canonical_name,
                url='http://example.com/image.raw.zst',
                md5sum=ImageMocksFixture.raw_md5sum,
                sha256=ImageMocksFixture.raw_sha256,
                timestamp=time.time(),
            )
            with iba.locked(ctx):
                model.ImageUpload.new(
                    ctx,
                    uuid='upload-uuid',
                    artifact_uuid='iba-uuid',
                    endpoint_name=endpoint.canonical_name,
                    providers=[provider.canonical_name],
                    canonical_name=image.canonical_name,
                    config_hash=image.config_hash,
                    timestamp=time.time(),
                    _state=model.ImageUpload.State.UPLOADING,
                    state_time=time.time(),
                )
                with iba.activeContext(ctx):
                    iba.state = iba.State.READY
        self.waitUntilSettled()
        pending_uploads = [
            u for u in self.launcher.image_upload_registry.getItems()
            if u.state == u.State.PENDING]
        self.assertEqual(0, len(pending_uploads))
        # If we reused the task, there should be two: the original
        # from test setup and our fake.
        self.assertEqual(2, len(list(endpoint._listImportSnapshotTasks())))

    @simple_layout('layouts/aws/nodepool-image-image.yaml',
                   enable_nodepool=True)
    @return_data(
        'build-debian-local-image',
        'refs/heads/master',
        AwsBaseTest.debian_return_data,
    )
    def test_aws_diskimage_image(self):
        self._test_diskimage()

    @simple_layout('layouts/aws/nodepool-image-image.yaml',
                   enable_nodepool=True)
    @return_data(
        'build-debian-local-image',
        'refs/heads/master',
        AwsBaseTest.debian_return_data,
    )
    def test_aws_diskimage_image_resume(self):
        self.addImageBuildEvent(
            'tenant-one',
            'review.example.com/org/common-config',
            'master',
            ['debian-local'],
        )
        self.waitUntilSettled()

        system_id = self.launcher.system.system_id
        image_format = 'raw'
        bucket_name = 'zuul'
        object_filename = 's3://zuul/image.raw'
        image_tags = [
            {'Key': 'zuul_system_id', 'Value': system_id},
            {'Key': 'zuul_upload_uuid', 'Value': 'upload-uuid'},
        ]

        task = self.fake_aws.import_image(
            DiskContainers=[{
                'Format': image_format,
                'UserBucket': {
                    'S3Bucket': bucket_name,
                    'S3Key': object_filename,
                },
            }],
            TagSpecifications=[{
                'ResourceType': 'import-image-task',
                'Tags': image_tags,
            }])
        image_id, snapshot_id = self.fake_aws.finish_import_image(task)

        provider = self.launcher._getProvider(
            'tenant-one', 'aws-us-east-1-main')
        endpoint = provider.getEndpoint()
        image = list(provider.images.values())[0]
        self.assertEqual('debian-local', image.name)
        # create an IBA and an upload
        with self.createZKContext(None) as ctx:
            # This starts with an unknown state, then
            # createImageUploads will set it to ready.
            iba = model.ImageBuildArtifact.new(
                ctx,
                uuid='iba-uuid',
                name=image.name,
                canonical_name=image.canonical_name,
                project_canonical_name=image.project_canonical_name,
                url='http://example.com/image.raw.zst',
                md5sum=ImageMocksFixture.raw_md5sum,
                sha256=ImageMocksFixture.raw_sha256,
                timestamp=time.time(),
            )
            with iba.locked(ctx):
                model.ImageUpload.new(
                    ctx,
                    uuid='upload-uuid',
                    artifact_uuid='iba-uuid',
                    endpoint_name=endpoint.canonical_name,
                    providers=[provider.canonical_name],
                    canonical_name=image.canonical_name,
                    config_hash=image.config_hash,
                    timestamp=time.time(),
                    _state=model.ImageUpload.State.UPLOADING,
                    state_time=time.time(),
                )
                with iba.activeContext(ctx):
                    iba.state = iba.State.READY
        self.waitUntilSettled()
        pending_uploads = [
            u for u in self.launcher.image_upload_registry.getItems()
            if u.state == u.State.PENDING]
        self.assertEqual(0, len(pending_uploads))
        # If we reused the task, there should be two: the original
        # from test setup and our fake.
        self.assertEqual(2, len(list(endpoint._listImportImageTasks())))

    @simple_layout('layouts/aws/nodepool-image-ebs-direct.yaml',
                   enable_nodepool=True)
    @return_data(
        'build-debian-local-image',
        'refs/heads/master',
        AwsBaseTest.debian_return_data,
    )
    def test_aws_diskimage_ebs_direct(self):
        self._test_diskimage()

    @simple_layout('layouts/aws/nodepool-image-snapshot.yaml',
                   enable_nodepool=True)
    @return_data(
        'build-debian-local-image',
        'refs/heads/master',
        AwsBaseTest.s3_debian_return_data,
    )
    def test_aws_diskimage_snapshot_import(self):
        self._test_diskimage()

    @simple_layout('layouts/aws/nodepool-image-image.yaml',
                   enable_nodepool=True)
    @return_data(
        'build-debian-local-image',
        'refs/heads/master',
        AwsBaseTest.s3_debian_return_data,
    )
    def test_aws_diskimage_image_import(self):
        self._test_diskimage()

    @simple_layout('layouts/aws/nodepool-image-ebs-direct.yaml',
                   enable_nodepool=True)
    @return_data(
        'build-debian-local-image',
        'refs/heads/master',
        AwsBaseTest.s3_debian_return_data,
    )
    def test_aws_diskimage_s3_download(self):
        # The ebs-direct method doesn't support an import from s3,
        # which means if we supply an s3 url, we will download it.
        bucket = self.s3.Bucket('zuul')
        bucket.put_object(Body=ImageMocksFixture.raw_body.encode('utf8'),
                          Key='image.raw')
        self._test_diskimage()

    @simple_layout('layouts/aws/nodepool-image-snapshot-region.yaml',
                   enable_nodepool=True)
    @return_data(
        'build-debian-local-image',
        'refs/heads/master',
        AwsBaseTest.s3_region_debian_return_data,
    )
    def test_aws_diskimage_s3_region_download(self):
        # The image in a bucket in a different region should be
        # downloaded without using a direct import.
        bucket = self.s3.Bucket('zuulwest')
        bucket.put_object(Body=ImageMocksFixture.raw_body.encode('utf8'),
                          Key='image.raw')
        self._test_diskimage()

    @simple_layout('layouts/aws/nodepool-image-copy.yaml',
                   enable_nodepool=True)
    @return_data(
        'build-debian-local-image',
        'refs/heads/master',
        AwsBaseTest.s3_debian_return_data,
    )
    def test_aws_diskimage_copy(self):
        # The image should be imported from s3 in one region and
        # copied to another.
        bucket = self.s3.Bucket('zuul')
        bucket.put_object(Body=ImageMocksFixture.raw_body.encode('utf8'),
                          Key='image.raw')
        self._test_diskimage(expected_uploads=2)
        self.assertEqual(1, len(self.copy_image_calls))

        self.launcher._runStats()
        self.assertReportedStat('zuul.uploads.state.ready',
                                value='2', kind='g')
        iname = 'review_example_com%2Forg%2Fcommon-config_debian-local'
        for region in ('us-east-1', 'us-west-1'):
            self.assertReportedStat(
                f'zuul.image.{iname}.upload.aws_aws-{region}.state.ready',
                value='1',
                kind='g')

    @simple_layout('layouts/aws/nodepool-image-copy.yaml',
                   enable_nodepool=True)
    @return_data(
        'build-debian-local-image',
        'refs/heads/master',
        AwsBaseTest.s3_debian_return_data,
    )
    def test_aws_diskimage_copy_resume(self):
        bucket = self.s3.Bucket('zuul')
        bucket.put_object(Body=ImageMocksFixture.raw_body.encode('utf8'),
                          Key='image.raw')
        self.addImageBuildEvent(
            'tenant-one',
            'review.example.com/org/common-config',
            'master',
            ['debian-local'],
        )
        self.waitUntilSettled()

        west_provider = self.launcher._getProvider(
            'tenant-one', 'aws-us-west-1-main')
        west_endpoint = west_provider.getEndpoint()

        for upload in self.launcher.image_upload_registry.getItems():
            if upload.endpoint_name == 'aws/aws-us-west-1':
                target_upload = upload

        # Tag the image since moto copy image doesn't
        system_id = self.launcher.system.system_id
        image_tags = [
            {'Key': 'zuul_system_id', 'Value': system_id},
            {'Key': 'zuul_upload_uuid', 'Value': target_upload.uuid},
        ]
        west_ec2_client = boto3.client('ec2', region_name='us-west-1')
        west_ec2_client.create_tags(
            Resources=[target_upload.external_id],
            Tags=image_tags)
        for _ in iterate_timeout(30, 'ami cache'):
            found = False
            for ami in west_endpoint._listAmis():
                if ami['Tags'] == image_tags:
                    found = True
                    break
            if found:
                break

        with self.createZKContext(None) as ctx:
            with target_upload.activeContext(ctx):
                self.log.debug("Resetting upload %s to pending",
                               target_upload)
                target_upload.state = target_upload.State.PENDING
                target_upload.external_id = None

        for _ in iterate_timeout(30, 'copy complete'):
            pending_uploads = [
                u for u in self.launcher.image_upload_registry.getItems()
                if u.state == u.State.PENDING]
            if not pending_uploads:
                break
        # If we reused the image, there should be one.
        self.assertEqual(1, len(self.copy_image_calls))

    @simple_layout('layouts/nodepool-multi-provider.yaml',
                   enable_nodepool=True)
    def test_aws_resource_cleanup(self):
        self.waitUntilSettled()
        self.launcher.cleanup_worker.INTERVAL = 1
        # This tests everything except the image imports
        # Start by setting up leaked resources
        system_id = self.launcher.system.system_id
        instance_tags = [
            {'Key': 'zuul_system_id', 'Value': system_id},
            {'Key': 'zuul_node_uuid', 'Value': '0000000042'},
        ]

        s3_tags = {
            'zuul_system_id': system_id,
            'zuul_upload_uuid': '0000000042',
        }

        reservation = self.ec2_client.run_instances(
            ImageId="ami-12c6146b", MinCount=1, MaxCount=1,
            BlockDeviceMappings=[{
                'DeviceName': '/dev/sda1',
                'Ebs': {
                    'VolumeSize': 80,
                    'DeleteOnTermination': False
                }
            }],
            TagSpecifications=[{
                'ResourceType': 'instance',
                'Tags': instance_tags
            }, {
                'ResourceType': 'volume',
                'Tags': instance_tags
            }]
        )
        instance_id = reservation['Instances'][0]['InstanceId']

        bucket = self.s3.Bucket('zuul')
        bucket.put_object(Body=b'hi',
                          Key='testimage',
                          Tagging=urllib.parse.urlencode(s3_tags))
        obj = self.s3.Object('zuul', 'testimage')
        # This effectively asserts the object exists
        self.s3_client.get_object_tagging(
            Bucket=obj.bucket_name, Key=obj.key)

        instance = self.ec2.Instance(instance_id)
        self.assertEqual(instance.state['Name'], 'running')

        volume_id = list(instance.volumes.all())[0].id
        volume = self.ec2.Volume(volume_id)
        self.assertEqual(volume.state, 'in-use')

        self.log.debug("Start cleanup worker")
        self.launcher.cleanup_worker.start()

        for _ in iterate_timeout(30, 'instance deletion'):
            instance = self.ec2.Instance(instance_id)
            if instance.state['Name'] == 'terminated':
                break
            time.sleep(1)

        for _ in iterate_timeout(30, 'volume deletion'):
            volume = self.ec2.Volume(volume_id)
            try:
                if volume.state == 'deleted':
                    break
            except botocore.exceptions.ClientError:
                # Probably not found
                break
            time.sleep(1)

        for _ in iterate_timeout(30, 'object deletion'):
            obj = self.s3.Object('zuul', 'testimage')
            try:
                self.s3_client.get_object_tagging(
                    Bucket=obj.bucket_name, Key=obj.key)
            except self.s3_client.exceptions.NoSuchKey:
                break
            time.sleep(1)

    @simple_layout('layouts/nodepool.yaml', enable_nodepool=True)
    def test_aws_resource_cleanup_import_snapshot(self):
        # This tests the import_snapshot path
        self.waitUntilSettled()
        self.launcher.cleanup_worker.INTERVAL = 1
        system_id = self.launcher.system.system_id

        # Start by setting up leaked resources
        image_tags = [
            {'Key': 'zuul_system_id', 'Value': system_id},
            {'Key': 'zuul_upload_uuid', 'Value': '0000000042'},
        ]

        task = self.fake_aws.import_snapshot(
            DiskContainer={
                'Format': 'ova',
                'UserBucket': {
                    'S3Bucket': 'zuul',
                    'S3Key': 'testfile',
                }
            },
            TagSpecifications=[{
                'ResourceType': 'import-snapshot-task',
                'Tags': image_tags,
            }])
        snapshot_id = self.fake_aws.finish_import_snapshot(task)

        register_response = self.ec2_client.register_image(
            Architecture='amd64',
            BlockDeviceMappings=[
                {
                    'DeviceName': '/dev/sda1',
                    'Ebs': {
                        'DeleteOnTermination': True,
                        'SnapshotId': snapshot_id,
                        'VolumeSize': 20,
                        'VolumeType': 'gp2',
                    },
                },
            ],
            RootDeviceName='/dev/sda1',
            VirtualizationType='hvm',
            Name='testimage',
        )
        image_id = register_response['ImageId']

        ami = self.ec2.Image(image_id)
        new_snapshot_id = ami.block_device_mappings[0]['Ebs']['SnapshotId']
        self.fake_aws.change_snapshot_id(task, new_snapshot_id)

        # Note that the resulting image and snapshot do not have tags
        # applied, so we test the automatic retagging methods in the
        # adapter.

        image = self.ec2.Image(image_id)
        self.assertEqual(image.state, 'available')

        snap = self.ec2.Snapshot(snapshot_id)
        self.assertEqual(snap.state, 'completed')

        # Now that the leaked resources exist, start the worker and
        # wait for it to clean them.
        self.log.debug("Start cleanup worker")
        self.launcher.cleanup_worker.start()

        for _ in iterate_timeout(30, 'ami deletion'):
            image = self.ec2.Image(image_id)
            try:
                # If this has a value the image was not deleted
                if image.state == 'available':
                    # Definitely not deleted yet
                    pass
            except AttributeError:
                # Per AWS API, a recently deleted image is empty and
                # looking at the state raises an AttributeFailure; see
                # https://github.com/boto/boto3/issues/2531.  The image
                # was deleted, so we continue on here
                break
            time.sleep(1)

        for _ in iterate_timeout(30, 'snapshot deletion'):
            snap = self.ec2.Snapshot(new_snapshot_id)
            try:
                if snap.state == 'deleted':
                    break
            except botocore.exceptions.ClientError:
                # Probably not found
                break
            time.sleep(1)

    @simple_layout('layouts/nodepool.yaml', enable_nodepool=True)
    def test_aws_resource_cleanup_import_image(self):
        # This tests the import_image path
        self.waitUntilSettled()
        self.launcher.cleanup_worker.INTERVAL = 1
        system_id = self.launcher.system.system_id

        # Start by setting up leaked resources
        image_tags = [
            {'Key': 'zuul_system_id', 'Value': system_id},
            {'Key': 'zuul_upload_uuid', 'Value': '0000000042'},
        ]

        # The image import path:
        task = self.fake_aws.import_image(
            DiskContainers=[{
                'Format': 'ova',
                'UserBucket': {
                    'S3Bucket': 'zuul',
                    'S3Key': 'testfile',
                }
            }],
            TagSpecifications=[{
                'ResourceType': 'import-image-task',
                'Tags': image_tags,
            }])
        image_id, snapshot_id = self.fake_aws.finish_import_image(task)

        # Note that the resulting image and snapshot do not have tags
        # applied, so we test the automatic retagging methods in the
        # adapter.

        image = self.ec2.Image(image_id)
        self.assertEqual(image.state, 'available')

        snap = self.ec2.Snapshot(snapshot_id)
        self.assertEqual(snap.state, 'completed')

        # Now that the leaked resources exist, start the provider and
        # wait for it to clean them.
        # Now that the leaked resources exist, start the worker and
        # wait for it to clean them.
        self.log.debug("Start cleanup worker")
        self.launcher.cleanup_worker.start()

        for _ in iterate_timeout(30, 'ami deletion'):
            image = self.ec2.Image(image_id)
            try:
                # If this has a value the image was not deleted
                if image.state == 'available':
                    # Definitely not deleted yet
                    pass
            except AttributeError:
                # Per AWS API, a recently deleted image is empty and
                # looking at the state raises an AttributeFailure; see
                # https://github.com/boto/boto3/issues/2531.  The image
                # was deleted, so we continue on here
                break
            time.sleep(1)

        for _ in iterate_timeout(30, 'snapshot deletion'):
            snap = self.ec2.Snapshot(snapshot_id)
            try:
                if snap.state == 'deleted':
                    break
            except botocore.exceptions.ClientError:
                # Probably not found
                break
            time.sleep(1)

    @simple_layout('layouts/nodepool-snapshot-expiration.yaml',
                   enable_nodepool=True)
    def test_snapshot_leak(self):
        # Ideally this would be in test_launcher, but it needs more of
        # AWS mocked out than we do there (the missing describe_*
        # paginators)
        ctx = self.createZKContext(None)
        request = self.requestNodes(['debian-normal'])

        # This is the executor's copy of the node
        node = model.ProviderNode.fromZK(
            ctx, path=model.ProviderNode._getPath(request.nodes[0]))

        with node.locked(ctx):
            request.delete(ctx)
            node.createSnapshot(ctx)
            with node.activeContext(ctx):
                self.log.debug("Set node to snapshot")
                node.setState(node.State.SNAPSHOT)

            for _ in iterate_timeout(60, "snapshot"):
                if node.snapshot.complete:
                    break
                node.snapshot.refresh(ctx)

            self.assertTrue(node.snapshot.complete)
            arn = node.snapshot.external_id
            with node.activeContext(ctx):
                self.log.debug("Set node to used")
                node.setState(node.State.USED)

        for _ in iterate_timeout(60, "node to be deleted"):
            try:
                node.refresh(ctx)
            except NoNodeError:
                break

        ec2 = boto3.resource('ec2', region_name='us-east-1')
        snapshot_id = arn.split('/')[1]
        snapshot = ec2.Snapshot(snapshot_id)
        tags = zuul.driver.aws.awsendpoint.tag_list_to_dict(
            snapshot.tags)
        # Assert that it has been re-tagged
        self.assertIn('zuul_system_id', tags)
        self.assertIn('zuul_expiration', tags)
        self.assertNotIn('zuul_upload_uuid', tags)

        self.launcher.cleanup_worker.INTERVAL = 1
        self.log.debug("Start cleanup worker")
        self.launcher.cleanup_worker.start()
        for _ in iterate_timeout(30, 'snapshot deletion'):
            snapshot = self.ec2.Snapshot(snapshot_id)
            try:
                if snapshot.state == 'deleted':
                    break
            except botocore.exceptions.ClientError:
                # Probably not found
                break
            time.sleep(1)

    @simple_layout('layouts/nodepool.yaml', enable_nodepool=True)
    def test_state_machines_instance(self):
        label_name = "debian-normal"
        provider_name = "aws-us-east-1-main"
        node_class = AwsProviderNode
        future_names = ['host_create_future', 'create_future']
        self._test_state_machines(label_name, provider_name,
                                  node_class, future_names)

    @simple_layout('layouts/nodepool.yaml', enable_nodepool=True)
    def test_state_machines_dedicated_host(self):
        label_name = "debian-dedicated"
        provider_name = "aws-us-east-1-main"
        node_class = AwsProviderNode
        future_names = ['host_create_future', 'create_future']
        self._test_state_machines(label_name, provider_name,
                                  node_class, future_names)

    @contextlib.contextmanager
    def _block_futures(self):
        with (mock.patch(
                'zuul.driver.aws.awsendpoint.AwsProviderEndpoint.'
                '_completeAllocateHost', return_value=None),
              mock.patch(
                'zuul.driver.aws.awsendpoint.AwsProviderEndpoint.'
                '_completeCreateInstance', return_value=None)):
            yield

    @simple_layout('layouts/aws/nodepool-multi-image.yaml',
                   enable_nodepool=True)
    def test_aws_multi_image(self):
        # Test that we can inherit aws attributes for both kinds of
        # images
        tenant = self.scheds.first.sched.abide.tenants.get("tenant-one")
        errors = tenant.layout.loading_errors
        self.assertEqual(len(errors), 0)

        images = tenant.layout.providers['aws-us-east-1-main'].images
        dl = images['debian-local']
        dc = images['debian-cloud']
        self.assertEqual('ebs-direct', dl.import_method)
        self.assertFalse(hasattr(dc, 'import_method'))
        self.waitUntilSettled()


class TestAwsSnapshot(AnsibleZuulTestCase, AwsBaseTest):
    tenant_config_file = 'config/snapshot/main.yaml'

    def lateSetUp(self):
        orig_getInstanceConfiguration = zuul.driver.aws.awsendpoint.\
            AwsProviderEndpoint._getInstanceConfiguration

        def getInstanceConfiguration(self, *args, **kw):
            args = orig_getInstanceConfiguration(self, *args, **kw)
            args['NetworkInterfaces'][0]['PrivateIpAddress'] = '127.0.0.1'
            return args
        self.patch(zuul.driver.aws.awsendpoint.AwsProviderEndpoint,
                   '_getInstanceConfiguration',
                   getInstanceConfiguration)

        orig_updateNodeFromInstance = zuul.launcher.server.Launcher.\
            _updateNodeFromInstance

        def _updateNodeFromInstance(self, node, instance):
            orig_updateNodeFromInstance(self, node, instance)
            node.connection_type = 'local'
        self.patch(zuul.launcher.server.Launcher,
                   '_updateNodeFromInstance',
                   _updateNodeFromInstance)

    def _waitForArtifacts(self, image_name, count):
        for _ in iterate_timeout(30, "artifacts to settle"):
            artifacts = self.launcher.image_build_registry.\
                getArtifactsForImage(image_name)
            if len(artifacts) == count:
                return artifacts

    def _waitForUploads(self, image_cname, count=None):
        for _ in iterate_timeout(60, "upload to complete"):
            uploads = self.launcher.image_upload_registry.getUploadsForImage(
                image_cname)
            pending = [u for u in uploads if u.external_id is None]
            if not pending:
                if count is None or count == len(uploads):
                    return uploads

    def test_snapshot_e2e(self):
        self.addImageBuildEvent(
            'tenant-one',
            'review.example.com/common-config',
            'master',
            ['debian-local'],
        )
        self.waitUntilSettled()
        image_cname = 'review.example.com%2Fcommon-config/debian-local'
        # We have an image, that's good enough for this test.
        artifacts = self._waitForArtifacts(image_cname, 1)
        self.assertTrue(artifacts[0].url.startswith(
            'arn:aws:ec2:us-east-1:123456789012:snapshot/snap-'))
        self._waitForUploads(image_cname, 1)

        paginator = self.ec2_client.get_paginator('describe_snapshots')
        snapshots = []
        for page in paginator.paginate(OwnerIds=['self']):
            snapshots.extend(page['Snapshots'])
        # Find our snapshot
        for snapshot in snapshots:
            tags = zuul.driver.aws.awsendpoint.tag_list_to_dict(
                snapshot['Tags'])
            if 'zuul_system_id' in tags:
                break
        # Assert that it has been re-tagged
        self.assertIn('zuul_system_id', tags)
        self.assertIn('zuul_upload_uuid', tags)
        self.assertNotIn('zuul_expiration', tags)



# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from docutils import nodes
from sphinx.domains import Domain

MODULE_URL = 'http://docs.ansible.com/ansible/latest/{module_name}_module.html'


def ansible_module_role(
        name, rawtext, text, lineno, inliner, options={}, content=[]):
    """Link to an upstream Ansible module.

    Returns 2 part tuple containing list of nodes to insert into the
    document and a list of system messages.  Both are allowed to be
    empty.

    :param name: The role name used in the document.
    :param rawtext: The entire markup snippet, with role.
    :param text: The text marked with the role.
    :param lineno: The line number where rawtext appears in the input.
    :param inliner: The inliner instance that called us.
    :param options: Directive options for customization.
    :param content: The directive content for customization.
    """
    node = nodes.reference(
        rawtext, "Ansible {module_name} module".format(module_name=text),
        refuri=MODULE_URL.format(module_name=text), **options)
    return ([node], [])


class AnsibleDomain(Domain):
    name = 'ansible'
    label = 'Ansible'

    roles = {
        'module': ansible_module_role,
    }


def setup(app):
    app.add_domain(AnsibleDomain)



# Copyright (C) 2023 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fixtures


class FakeSocket:
    def __init__(self, *args):
        self.blocking = True
        self.fd = 1

    def setblocking(self, b):
        self.blocking = b

    def getsockopt(self, level, optname):
        return None

    def connect(self, addr):
        if not self.blocking:
            raise BlockingIOError()
        raise Exception("blocking connect attempted")

    def fileno(self):
        return self.fd


class FakePoll:
    _INIT_FAIL = False

    def __init__(self):
        self.fds = []
        self._fail = self._INIT_FAIL

    def register(self, fd, bitmap):
        self.fds.append(fd)

    def unregister(self, fd):
        if fd in self.fds:
            self.fds.remove(fd)

    def poll(self, timeout=None):
        if self._fail:
            return []
        fds = self.fds[:]
        self.fds = [f for f in fds if not isinstance(f, FakeSocket)]
        fds = [f.fileno() if hasattr(f, 'fileno') else f for f in fds]
        return [(f, 0) for f in fds]


class Dummy:
    pass


class FakeKey:
    def get_name(self):
        return 'fake key'

    def get_base64(self):
        return 'fake base64'


class FakeTransport:
    _INIT_ACTIVE = True
    _INIT_FAIL = False

    def __init__(self, *args, **kw):
        self.active = self._INIT_ACTIVE
        self._fail = self._INIT_FAIL

    def start_client(self, event=None, timeout=None):
        if not self._fail:
            event.set()

    def get_security_options(self):
        ret = Dummy()
        ret.key_types = ['rsa']
        return ret

    def get_remote_server_key(self):
        return FakeKey()

    def get_exception(self):
        return Exception("Fake ssh error")


class NodescanFixture(fixtures.Fixture):
    def __init__(self, transport_active=True, transport_fail=False,
                 poll_fail=False):
        super().__init__()
        self._transport_active = transport_active
        self._transport_fail = transport_fail
        self._poll_fail = poll_fail

    def _setUp(self):
        self.useFixture(fixtures.MonkeyPatch(
            'zuul.launcher.server.NodescanRequest._socket_class',
            FakeSocket))
        self.useFixture(fixtures.MonkeyPatch(
            'zuul.launcher.server.NodescanWorker._poll_class',
            FakePoll))
        self.useFixture(fixtures.MonkeyPatch(
            'paramiko.transport.Transport',
            FakeTransport))
        self.useFixture(fixtures.MonkeyPatch(
            'tests.fake_nodescan.FakeTransport._INIT_ACTIVE',
            self._transport_active))
        self.useFixture(fixtures.MonkeyPatch(
            'tests.fake_nodescan.FakeTransport._INIT_FAIL',
            self._transport_fail))
        self.useFixture(fixtures.MonkeyPatch(
            'tests.fake_nodescan.FakePoll._INIT_FAIL',
            self._poll_fail))



# Copyright 2022 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import time

from tests.base import iterate_timeout, ZuulTestCase

import zuul.lib.tracing as tracing
from opentelemetry import trace


def attributes_to_dict(attrlist):
    ret = {}
    for attr in attrlist:
        if attr.value.string_value:
            ret[attr.key] = attr.value.string_value
        else:
            ret[attr.key] = [v.string_value
                             for v in attr.value.array_value.values]
    return ret


class TestTracing(ZuulTestCase):
    config_file = 'zuul-tracing.conf'
    tenant_config_file = "config/single-tenant/main.yaml"

    def _waitForSpans(self, *span_names, timeout=60,):
        for _ in iterate_timeout(timeout, "requests to arrive"):
            test_requests = [
                r for r in self.otlp.requests
                if r.resource_spans[0].scope_spans[0].spans[0].name
                in span_names
            ]
            if len(test_requests) == len(span_names):
                return test_requests

    def getSpan(self, name):
        for req in self.otlp.requests:
            span = req.resource_spans[0].scope_spans[0].spans[0]
            if span.name == name:
                return span

    def test_tracing_api(self):
        tracer = trace.get_tracer("zuul")

        # We have a lot of timestamps stored as floats, so make sure
        # our root span is a ZuulSpan that can handle that input.
        span_info = tracing.startSavedSpan('parent-trace',
                                           start_time=time.time(),
                                           attributes={'startattr': 'bar'},
                                           include_attributes=True)

        # Simulate a reconstructed root span
        span = tracing.restoreSpan(span_info)

        # Within the root span, use the more typical OpenTelemetry
        # context manager api.
        with trace.use_span(span):
            with tracer.start_span('child1-trace') as child1_span:
                link = trace.Link(child1_span.context,
                                  attributes={'relationship': 'prev'})

        # Make sure that we can manually start and stop a child span,
        # and that it is a ZuulSpan as well.
        with trace.use_span(span):
            child = tracer.start_span('child2-trace', start_time=time.time(),
                                      links=[link])
            child.end(end_time=time.time())

        # Make sure that we can start a child span from a span
        # context and not a full span:
        span_context = tracing.getSpanContext(span)
        with tracing.startSpanInContext(span_context, 'child3-trace') as child:
            child.end(end_time=time.time())

        # End our root span manually.
        tracing.endSavedSpan(span_info, end_time=time.time(),
                             attributes={'endattr': 'baz'})

        test_requests = self._waitForSpans(
            "parent-trace", "child1-trace", "child2-trace", "child3-trace")

        req1 = test_requests[0]
        self.log.debug("Received:\n%s", req1)
        attrs = attributes_to_dict(req1.resource_spans[0].resource.attributes)
        self.assertEqual({"service.name": "zuultest"}, attrs)
        self.assertEqual("zuul",
                         req1.resource_spans[0].scope_spans[0].scope.name)
        span1 = req1.resource_spans[0].scope_spans[0].spans[0]
        self.assertEqual("child1-trace", span1.name)

        req2 = test_requests[1]
        self.log.debug("Received:\n%s", req2)
        span2 = req2.resource_spans[0].scope_spans[0].spans[0]
        self.assertEqual("child2-trace", span2.name)
        self.assertEqual(span2.links[0].span_id, span1.span_id)
        attrs = attributes_to_dict(span2.links[0].attributes)
        self.assertEqual({"relationship": "prev"}, attrs)

        req3 = test_requests[2]
        self.log.debug("Received:\n%s", req3)
        span3 = req3.resource_spans[0].scope_spans[0].spans[0]
        self.assertEqual("child3-trace", span3.name)

        req4 = test_requests[3]
        self.log.debug("Received:\n%s", req4)
        span4 = req4.resource_spans[0].scope_spans[0].spans[0]
        self.assertEqual("parent-trace", span4.name)
        attrs = attributes_to_dict(span4.attributes)
        self.assertEqual({"startattr": "bar",
                          "endattr": "baz"}, attrs)

        self.assertEqual(span1.trace_id, span4.trace_id)
        self.assertEqual(span2.trace_id, span4.trace_id)
        self.assertEqual(span3.trace_id, span4.trace_id)

    def test_tracing_api_null(self):
        tracer = trace.get_tracer("zuul")

        # Test that restoring spans and span contexts works with
        # null values.

        span_info = None
        # Simulate a reconstructed root span from a null value
        span = tracing.restoreSpan(span_info)

        # Within the root span, use the more typical OpenTelemetry
        # context manager api.
        with trace.use_span(span):
            with tracer.start_span('child1-trace') as child1_span:
                link = trace.Link(child1_span.context,
                                  attributes={'relationship': 'prev'})

        # Make sure that we can manually start and stop a child span,
        # and that it is a ZuulSpan as well.
        with trace.use_span(span):
            child = tracer.start_span('child2-trace', start_time=time.time(),
                                      links=[link])
            child.end(end_time=time.time())

        # Make sure that we can start a child span from a null span
        # context:
        span_context = None
        with tracing.startSpanInContext(span_context, 'child3-trace') as child:
            child.end(end_time=time.time())

        # End our root span manually.
        span.end(end_time=time.time())

        test_requests = self._waitForSpans(
            "child1-trace", "child2-trace", "child3-trace")

        req1 = test_requests[0]
        self.log.debug("Received:\n%s", req1)
        attrs = attributes_to_dict(req1.resource_spans[0].resource.attributes)
        self.assertEqual({"service.name": "zuultest"}, attrs)
        self.assertEqual("zuul",
                         req1.resource_spans[0].scope_spans[0].scope.name)
        span1 = req1.resource_spans[0].scope_spans[0].spans[0]
        self.assertEqual("child1-trace", span1.name)

        req2 = test_requests[1]
        self.log.debug("Received:\n%s", req2)
        span2 = req2.resource_spans[0].scope_spans[0].spans[0]
        self.assertEqual("child2-trace", span2.name)
        self.assertEqual(span2.links[0].span_id, span1.span_id)
        attrs = attributes_to_dict(span2.links[0].attributes)
        self.assertEqual({"relationship": "prev"}, attrs)

        req3 = test_requests[2]
        self.log.debug("Received:\n%s", req3)
        span3 = req3.resource_spans[0].scope_spans[0].spans[0]
        self.assertEqual("child3-trace", span3.name)

        self.assertNotEqual(span1.trace_id, span2.trace_id)
        self.assertNotEqual(span2.trace_id, span3.trace_id)
        self.assertNotEqual(span1.trace_id, span3.trace_id)

    def test_tracing(self):
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        A.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        for _ in iterate_timeout(60, "request to arrive"):
            if len(self.otlp.requests) >= 2:
                break

        buildset = self.getSpan('BuildSet')
        self.log.debug("Received:\n%s", buildset)
        item = self.getSpan('QueueItem')
        self.log.debug("Received:\n%s", item)
        merge_job = self.getSpan('Merge')
        self.log.debug("Received:\n%s", merge_job)
        node_request = self.getSpan('RequestNodes')
        self.log.debug("Received:\n%s", node_request)
        build = self.getSpan('Build')
        self.log.debug("Received:\n%s", build)
        jobexec = self.getSpan('JobExecution')
        self.log.debug("Received:\n%s", jobexec)

        build_update_lock = self.getSpan('BuildRepoUpdateLock')
        self.log.debug("Received:\n%s", build_update_lock)
        build_update = self.getSpan('BuildRepoUpdate')
        self.log.debug("Received:\n%s", build_update)
        build_clone = self.getSpan('BuildCloneRepo')
        self.log.debug("Received:\n%s", build_clone)
        build_merge = self.getSpan('BuildMergeChanges')
        self.log.debug("Received:\n%s", build_merge)
        # BuildSetRepoState not exercised in this test
        build_checkout = self.getSpan('BuildCheckout')
        self.log.debug("Received:\n%s", build_checkout)

        self.assertEqual(item.trace_id, buildset.trace_id)
        self.assertEqual(item.trace_id, node_request.trace_id)
        self.assertEqual(item.trace_id, build.trace_id)
        self.assertNotEqual(item.span_id, jobexec.span_id)
        self.assertTrue(buildset.start_time_unix_nano >=
                        item.start_time_unix_nano)
        self.assertTrue(buildset.end_time_unix_nano <=
                        item.end_time_unix_nano)
        self.assertTrue(merge_job.start_time_unix_nano >=
                        buildset.start_time_unix_nano)
        self.assertTrue(merge_job.end_time_unix_nano <=
                        buildset.end_time_unix_nano)
        self.assertEqual(jobexec.parent_span_id,
                         build.span_id)
        self.assertEqual(node_request.parent_span_id,
                         buildset.span_id)
        self.assertEqual(build.parent_span_id,
                         buildset.span_id)
        self.assertEqual(merge_job.parent_span_id,
                         buildset.span_id)
        self.assertEqual(buildset.parent_span_id,
                         item.span_id)

        self.assertEqual(build_update_lock.parent_span_id,
                         jobexec.span_id)
        self.assertEqual(build_update.parent_span_id,
                         jobexec.span_id)
        self.assertEqual(build_clone.parent_span_id,
                         jobexec.span_id)
        self.assertEqual(build_merge.parent_span_id,
                         jobexec.span_id)
        self.assertEqual(build_checkout.parent_span_id,
                         jobexec.span_id)

        item_attrs = attributes_to_dict(item.attributes)
        self.assertTrue(item_attrs['ref_number'] == ["1"])
        self.assertTrue(item_attrs['ref_patchset'] == ["1"])
        self.assertTrue('zuul_event_id' in item_attrs)

    def test_post(self):
        "Test that post jobs run"
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        A.setMerged()
        self.fake_gerrit.addEvent(A.getRefUpdatedEvent())
        self.waitUntilSettled()

        self.assertHistory([
            dict(name='project-post', result='SUCCESS',
                 ref='refs/heads/master'),
        ], ordered=False)

        job_names = [x.name for x in self.history]
        self.assertEqual(len(self.history), 1)
        self.assertIn('project-post', job_names)

    def test_timer(self):
        self.executor_server.hold_jobs_in_build = True
        self.commitConfigUpdate('common-config', 'layouts/timer.yaml')
        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))
        self.waitUntilSettled()

        for _ in iterate_timeout(30, 'Wait for a build on hold'):
            if len(self.builds) > 0:
                break
        self.waitUntilSettled()

        self.commitConfigUpdate('common-config',
                                'layouts/no-timer.yaml')
        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))
        self.waitUntilSettled()
        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()
        self.assertHistory([
            dict(name='project-bitrot', result='SUCCESS',
                 ref='refs/heads/master'),
        ], ordered=False)



# Copyright 2014 Puppet Labs Inc
# Copyright 2023 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re
import urllib
import logging
import time
import voluptuous as v

from zuul.source import BaseSource
from zuul.model import MERGER_MAP, Project
from zuul.driver.github.githubmodel import GithubRefFilter
from zuul.driver.util import scalar_or_list
from zuul.zk.change_cache import ChangeKey
from zuul.zk.components import COMPONENT_REGISTRY


class GithubSource(BaseSource):
    name = 'github'
    log = logging.getLogger("zuul.source.GithubSource")

    def __init__(self, driver, connection, config=None):
        hostname = connection.canonical_hostname
        super(GithubSource, self).__init__(driver, connection,
                                           hostname, config)

    def getRefSha(self, project, ref):
        """Return a sha for a given project ref."""
        raise NotImplementedError()

    def waitForRefSha(self, project, ref, old_sha=''):
        """Block until a ref shows up in a given project."""
        raise NotImplementedError()

    def isMerged(self, change, head=None):
        """Determine if change is merged."""
        if not change.number:
            # Not a pull request, considering merged.
            return True
        # We don't need to perform another query because the API call
        # to perform the merge will ensure this is updated.
        if change.is_merged:
            return change.merge_commit_sha or True
        return False

    def canMerge(self, change, allow_needs, event=None, allow_refresh=False):
        """Determine if change can merge."""

        if not change.number:
            # Not a pull request, considering merged.
            return True
        return self.connection.canMerge(
            change, allow_needs, event=event, allow_refresh=allow_refresh
        )

    def postConfig(self):
        """Called after configuration has been processed."""
        pass

    def getChangeKey(self, event):
        connection_name = self.connection.connection_name
        if event.change_number:
            return ChangeKey(connection_name, event.project_name,
                             'PullRequest',
                             str(event.change_number),
                             str(event.patch_number))
        revision = f'{event.oldrev}..{event.newrev}'
        if event.ref and event.ref.startswith('refs/tags/'):
            tag = event.ref[len('refs/tags/'):]
            return ChangeKey(connection_name, event.project_name,
                             'Tag', tag, revision)
        if event.ref and event.ref.startswith('refs/heads/'):
            branch = event.ref[len('refs/heads/'):]
            return ChangeKey(connection_name, event.project_name,
                             'Branch', branch, revision)
        if event.ref:
            return ChangeKey(connection_name, event.project_name,
                             'Ref', event.ref, revision)

        self.log.warning("Unable to format change key for %s" % (self,))

    def getChange(self, change_key, refresh=False, event=None):
        return self.connection.getChange(change_key, refresh=refresh,
                                         event=event)

    change_re = re.compile(r"/(.*?)/(.*?)/pull/(\d+)[\w]*")

    def getChangeByURL(self, url, event):
        try:
            parsed = urllib.parse.urlparse(url)
        except ValueError:
            return None
        m = self.change_re.match(parsed.path)
        if not m:
            return None
        org = m.group(1)
        proj = m.group(2)
        try:
            num = int(m.group(3))
        except ValueError:
            return None
        pull, pr_obj = self.connection.getPull(
            '%s/%s' % (org, proj), int(num), event=event)
        if not pull:
            return None
        proj = pull.get('base').get('repo').get('full_name')
        change_key = ChangeKey(self.connection.connection_name, proj,
                               'PullRequest',
                               str(num),
                               pull.get('head').get('sha'))
        change = self.connection._getChange(change_key, event=event)
        return change

    def getChangesDependingOn(self, change, projects, tenant):
        return self.connection.getChangesDependingOn(change, projects, tenant)

    def getCachedChanges(self):
        return list(self.connection._change_cache)

    def getProject(self, name):
        p = self.connection.getProject(name)
        if not p:
            p = Project(name, self)
            self.connection.addProject(p)
        return p

    def getProjectBranches(self, project, tenant, min_ltime=-1):
        return self.connection.getProjectBranches(project, tenant, min_ltime)

    def getProjectMergeModes(self, project, tenant, min_ltime=-1):
        return self.connection.getProjectMergeModes(project, tenant, min_ltime)

    def getProjectDefaultBranch(self, project, tenant, min_ltime=-1):
        # We have to return something here, so try to get it from the
        # cache, and if that fails, return the Zuul default.
        try:
            default_branch = self.connection.getProjectDefaultBranch(
                project, tenant, min_ltime)
        except Exception:
            default_branch = None
        if default_branch is None:
            return super().getProjectDefaultBranch(project, tenant, min_ltime)
        return default_branch

    def getProjectBranchSha(self, project, branch_name):
        return self.connection.getProjectBranchSha(project, branch_name)

    def getProjectBranchCacheLtime(self):
        return self.connection._branch_cache.ltime

    def getProjectOpenChanges(self, project):
        """Get the open changes for a project."""
        raise NotImplementedError()

    def getProjectDefaultMergeMode(self, project, valid_modes=None):
        if COMPONENT_REGISTRY.model_api < 18:
            return 'merge'
        github_version = self.connection._github_client_manager._github_version
        if github_version and github_version < (3, 8):
            merge_mode = 'merge-recursive'
        else:
            merge_mode = 'merge-ort'
        try:
            if valid_modes and MERGER_MAP[merge_mode] in valid_modes:
                return merge_mode
        except Exception:
            self.log.exception("Failed to get valid merge modes:")
        return 'merge'

    def updateChange(self, change, history=None):
        """Update information for a change."""
        raise NotImplementedError()

    def getGitUrl(self, project):
        """Get the git url for a project."""
        return self.connection.getGitUrl(project)

    def getRetryTimeout(self, project):
        """Get the retry timeout for a project."""
        return self.connection.repo_retry_timeout

    def getGitwebUrl(self, project, sha=None):
        """Get the git-web url for a project."""
        return self.connection.getGitwebUrl(project, sha)

    def _ghTimestampToDate(self, timestamp):
        return time.strptime(timestamp, '%Y-%m-%dT%H:%M:%SZ')

    def getRequireFilters(self, config, parse_context):
        f = GithubRefFilter.requiresFromConfig(
            self.connection.connection_name,
            config)
        return [f]

    def getRejectFilters(self, config, parse_context):
        f = GithubRefFilter.rejectFromConfig(
            self.connection.connection_name,
            config)
        return [f]

    def getRefForChange(self, change):
        return "refs/pull/%s/head" % change

    def setChangeAttributes(self, change, **attrs):
        return self.connection.updateChangeAttributes(change, **attrs)


review = v.Schema({'username': str,
                   'email': str,
                   'older-than': str,
                   'newer-than': str,
                   'type': str,
                   'permission': v.Any('read', 'write', 'admin'),
                   })


def getRequireSchema():
    require = {'status': scalar_or_list(str),
               'review': scalar_or_list(review),
               'open': bool,
               'merged': bool,
               'current-patchset': bool,
               'draft': bool,
               'label': scalar_or_list(str)}
    return require


def getRejectSchema():
    reject = {'status': scalar_or_list(str),
              'review': scalar_or_list(review),
              'open': bool,
              'merged': bool,
              'current-patchset': bool,
              'draft': bool,
              'label': scalar_or_list(str)}
    return reject



# This file is derived from the Kazoo project
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from kazoo.exceptions import (
    AuthFailedError,
    EXCEPTIONS,
    NoNodeError,
)
from kazoo.loggingsupport import BLATHER
from kazoo.protocol.connection import (
    ConnectionHandler,
    CREATED_EVENT,
    DELETED_EVENT,
    CHANGED_EVENT,
    CHILD_EVENT,
    CLOSE_RESPONSE,
    WATCH_XID,
    PING_XID,
    AUTH_XID,
)
from kazoo.protocol.serialization import (
    Close,
    Exists,
    GetChildren,
    GetChildren2,
    Transaction,
    Watch,
)
from kazoo.protocol.states import (
    Callback,
    KeeperState,
    EVENT_TYPE_MAP,
)

from zuul.zk.vendor.states import (
    AddWatchMode,
    WatcherType,
    WatchedEvent,
)
from zuul.zk.vendor.serialization import (
    AddWatch,
    RemoveWatches,
)


class ZuulConnectionHandler(ConnectionHandler):

    def _find_persistent_recursive_watchers(self, path):
        parts = path.split("/")
        watchers = []
        for count in range(len(parts)):
            candidate = "/".join(parts[: count + 1])
            if not candidate:
                continue
            watchers.extend(
                self.client._persistent_recursive_watchers.get(candidate, [])
            )
        return watchers

    def _read_watch_event(self, buffer, offset, zxid):
        client = self.client
        watch, offset = Watch.deserialize(buffer, offset)
        path = watch.path

        self.logger.debug("Received EVENT: %s", watch)

        watchers = []

        if watch.type in (CREATED_EVENT, CHANGED_EVENT):
            watchers.extend(client._data_watchers.pop(path, []))
            watchers.extend(client._persistent_watchers.get(path, []))
            watchers.extend(self._find_persistent_recursive_watchers(path))
        elif watch.type == DELETED_EVENT:
            watchers.extend(client._data_watchers.pop(path, []))
            watchers.extend(client._child_watchers.pop(path, []))
            watchers.extend(client._persistent_watchers.get(path, []))
            watchers.extend(self._find_persistent_recursive_watchers(path))
        elif watch.type == CHILD_EVENT:
            watchers.extend(client._child_watchers.pop(path, []))
        else:
            self.logger.warn("Received unknown event %r", watch.type)
            return

        # Strip the chroot if needed
        path = client.unchroot(path)
        ev = WatchedEvent(EVENT_TYPE_MAP[watch.type], client._state, path,
                          zxid)

        # Last check to ignore watches if we've been stopped
        if client._stopped.is_set():
            return

        # Dump the watchers to the watch thread
        for watch in watchers:
            client.handler.dispatch_callback(Callback("watch", watch, (ev,)))

    def _read_response(self, header, buffer, offset):
        client = self.client
        request, async_object, xid = client._pending.popleft()
        if header.zxid and header.zxid > 0:
            client.last_zxid = header.zxid
        if header.xid != xid:
            exc = RuntimeError(
                "xids do not match, expected %r " "received %r",
                xid,
                header.xid,
            )
            async_object.set_exception(exc)
            raise exc

        # Determine if its an exists request and a no node error
        exists_error = (
            header.err == NoNodeError.code and request.type == Exists.type
        )

        # Set the exception if its not an exists error
        if header.err and not exists_error:
            callback_exception = EXCEPTIONS[header.err]()
            self.logger.debug(
                "Received error(xid=%s) %r", xid, callback_exception
            )
            if async_object:
                async_object.set_exception(callback_exception)
        elif request and async_object:
            if exists_error:
                # It's a NoNodeError, which is fine for an exists
                # request
                async_object.set(None)
            else:
                try:
                    response = request.deserialize(buffer, offset)
                except Exception as exc:
                    self.logger.exception(
                        "Exception raised during deserialization "
                        "of request: %s",
                        request,
                    )
                    async_object.set_exception(exc)
                    return
                self.logger.debug(
                    "Received response(xid=%s): %r", xid, response
                )

                # We special case a Transaction as we have to unchroot things
                if request.type == Transaction.type:
                    response = Transaction.unchroot(client, response)

                async_object.set(response)

            # Determine if watchers should be registered or unregistered
            if not client._stopped.is_set():
                watcher = getattr(request, "watcher", None)
                if watcher:
                    if isinstance(request, AddWatch):
                        if request.mode == AddWatchMode.PERSISTENT:
                            client._persistent_watchers[request.path].add(
                                watcher
                            )
                        elif request.mode == AddWatchMode.PERSISTENT_RECURSIVE:
                            client._persistent_recursive_watchers[
                                request.path
                            ].add(watcher)
                    elif isinstance(request, (GetChildren, GetChildren2)):
                        client._child_watchers[request.path].add(watcher)
                    else:
                        client._data_watchers[request.path].add(watcher)
                if isinstance(request, RemoveWatches):
                    if request.watcher_type == WatcherType.CHILDREN:
                        client._child_watchers.pop(request.path, None)
                    elif request.watcher_type == WatcherType.DATA:
                        client._data_watchers.pop(request.path, None)
                    elif request.watcher_type == WatcherType.ANY:
                        client._child_watchers.pop(request.path, None)
                        client._data_watchers.pop(request.path, None)
                        client._persistent_watchers.pop(request.path, None)
                        client._persistent_recursive_watchers.pop(
                            request.path, None
                        )

        if isinstance(request, Close):
            self.logger.log(BLATHER, "Read close response")
            return CLOSE_RESPONSE

    def _read_socket(self, read_timeout):
        """Called when there's something to read on the socket"""
        client = self.client

        header, buffer, offset = self._read_header(read_timeout)
        if header.xid == PING_XID:
            self.logger.log(BLATHER, "Received Ping")
            self.ping_outstanding.clear()
        elif header.xid == AUTH_XID:
            self.logger.log(BLATHER, "Received AUTH")

            request, async_object, xid = client._pending.popleft()
            if header.err:
                async_object.set_exception(AuthFailedError())
                client._session_callback(KeeperState.AUTH_FAILED)
            else:
                async_object.set(True)
        elif header.xid == WATCH_XID:
            self._read_watch_event(buffer, offset, header.zxid)
        else:
            self.logger.log(BLATHER, "Reading for header %r", header)

            return self._read_response(header, buffer, offset)



# Copyright 2024-2025 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import logging

from zuul.connection import BaseConnection


class AzureConnection(BaseConnection):
    driver_name = 'azure'
    log = logging.getLogger("zuul.AzureConnection")

    def __init__(self, driver, connection_name, connection_config):
        super().__init__(driver, connection_name, connection_config)

        # Users can provide credentials directly in zuul.conf, a
        # credentials file, or a federated token file.  We will check
        # those places in that order.
        self.shared_credentials_file = self.connection_config.get(
            'shared_credentials_file')
        self.federated_token_file = self.connection_config.get(
            'federated_token_file')

        credential = {}
        if self.shared_credentials_file:
            with open(self.shared_credentials_file, encoding="utf-8") as f:
                credential = json.load(f)

        self.tenant_id = self.connection_config.get(
            'tenant_id', credential.get('tenantId'))
        self.client_id = self.connection_config.get(
            'client_id', credential.get('clientId'))
        self.client_secret = self.connection_config.get(
            'client_secret', credential.get('clientSecret'))
        self.subscription_id = self.connection_config.get(
            'subscription_id', credential.get('subscriptionId'))

        # Rate limit: requests/second
        self.rate = self.connection_config.get('rate', 2)



# Copyright 2012 Hewlett-Packard Development Company, L.P.
# Copyright 2013 OpenStack Foundation
# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import fcntl
import grp
import logging
import os
import os.path
import psutil
import pwd
import shlex
import subprocess
import threading
import re
import struct

from typing import Dict, List  # noqa

from zuul.driver import (Driver, WrapperInterface)
from zuul.execution_context import BaseExecutionContext

# Increase the OOM badness score by about 20% of the available RAM.
# This should be sufficient to make child processes the target of the
# oom killer.
OOM_SCORE_ADJ = 200


class WrappedPopen(object):
    def __init__(self, command, fds):
        self.command = command
        self.fds = fds

    def __call__(self, args, *sub_args, **kwargs):
        try:
            args = self.command + args
            pass_fds = list(kwargs.get('pass_fds', []))
            for fd in self.fds:
                if fd not in pass_fds:
                    pass_fds.append(fd)
            kwargs['pass_fds'] = pass_fds
            proc = psutil.Popen(args, *sub_args, **kwargs)
        finally:
            self.__del__()
        return proc

    def __del__(self):
        for fd in self.fds:
            try:
                os.close(fd)
            except OSError:
                pass
        self.fds = []


class BubblewrapExecutionContext(BaseExecutionContext):
    log = logging.getLogger("zuul.BubblewrapExecutionContext")

    def __init__(self, bwrap_command, ro_paths, rw_paths, secrets):
        self.bwrap_command = bwrap_command
        self.mounts_map = {'ro': ro_paths, 'rw': rw_paths}
        self.secrets = secrets

    def getNamespacePids(self, proc):
        # Given a Popen object (proc), return the namespace and a list
        # of host-side process ids in proc's namespace.
        ps = subprocess.Popen(
            ['ps', '-axo', 'pidns,pid,ppid'],
            stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)
        pid_to_child_list = {}
        ns_to_pid_list = {}
        pid_to_ns_map = {}
        for line in ps.stdout:
            try:
                (pidns, pid, ppid) = map(int, line.rstrip().split())
            except ValueError:
                continue
            pid_to_child_list.setdefault(ppid, []).append(pid)
            ns_to_pid_list.setdefault(pidns, []).append(pid)
            pid_to_ns_map[pid] = pidns
        for child in pid_to_child_list.get(proc.pid, []):
            ns = pid_to_ns_map.get(child)
            if ns is not None:
                return ns, ns_to_pid_list.get(ns)
        return None, []

    def startPipeWriter(self, pipe, data):
        # In case we have a large amount of data to write through a
        # pipe, spawn a thread to handle the writes.
        t = threading.Thread(target=self._writer, args=(pipe, data))
        t.daemon = True
        t.start()

    def _writer(self, pipe, data):
        os.write(pipe, data)
        os.close(pipe)

    def setpag(self):
        # If we are on a system with AFS, ensure that each playbook
        # invocation ends up in its own PAG.
        # http://asa.scripts.mit.edu/trac/attachment/ticket/145/setpag.txt#L315
        if os.path.exists("/proc/fs/openafs/afs_ioctl"):
            f = os.open("/proc/fs/openafs/afs_ioctl", os.O_RDONLY)
            # 0x40084301 is the result of _IOW('C', 1, void *) which originates
            # from OpenAFS' VIOC_SYSCALL defined at:
            # https://github.com/openafs/openafs/blob/
            # 630d423897e5fffed1873aa9d12c4e74a8481041/
            # src/config/afs_args.h#L258
            # The 21 at the end of the struct is AFSCALL_SETPAG which
            # is defined at:
            # https://github.com/openafs/openafs/blob/
            # 630d423897e5fffed1873aa9d12c4e74a8481041/
            # src/config/afs_args.h#L97
            fcntl.ioctl(f, 0x40084301, struct.pack("lllll", 0, 0, 0, 0, 21))
            os.close(f)

    def getPopen(self, **kwargs):
        self.setpag()
        bwrap_command = list(self.bwrap_command)
        for mount_type in ('ro', 'rw'):
            bind_arg = '--ro-bind' if mount_type == 'ro' else '--bind'
            for bind in self.mounts_map[mount_type]:
                bwrap_command.extend([bind_arg, bind, bind])

        # A list of file descriptors which must be held open so that
        # bwrap may read from them.
        read_fds = []
        # Need users and groups
        uid = os.getuid()
        passwd = list(pwd.getpwuid(uid))
        # Replace our user's actual home directory with the work dir.
        passwd = passwd[:5] + [kwargs['work_dir']] + passwd[6:]
        passwd_bytes = b':'.join(
            ['{}'.format(x).encode('utf8') for x in passwd])
        (passwd_r, passwd_w) = os.pipe()
        os.write(passwd_w, passwd_bytes)
        os.write(passwd_w, b'\n')
        os.close(passwd_w)
        read_fds.append(passwd_r)

        gid = os.getgid()
        group = grp.getgrgid(gid)
        group_bytes = b':'.join(
            ['{}'.format(x).encode('utf8') for x in group])
        group_r, group_w = os.pipe()
        os.write(group_w, group_bytes)
        os.write(group_w, b'\n')
        os.close(group_w)
        read_fds.append(group_r)

        # Create a tmpfs for each directory which holds secrets, and
        # tell bubblewrap to write the contents to a file therein.
        secret_dirs = set()
        for fn, content in self.secrets.items():
            secret_dir = os.path.dirname(fn)
            if secret_dir not in secret_dirs:
                bwrap_command.extend(['--tmpfs', secret_dir])
                secret_dirs.add(secret_dir)
            secret_r, secret_w = os.pipe()
            self.startPipeWriter(secret_w, content.encode('utf8'))
            bwrap_command.extend(['--file', str(secret_r), fn])
            read_fds.append(secret_r)

        kwargs = dict(kwargs)  # Don't update passed in dict
        kwargs['uid'] = uid
        kwargs['gid'] = gid
        kwargs['uid_fd'] = passwd_r
        kwargs['gid_fd'] = group_r
        command = [x.format(**kwargs) for x in bwrap_command]

        self.log.debug("Bubblewrap command: %s",
                       " ".join(shlex.quote(c) for c in command))

        wrapped_popen = WrappedPopen(command, read_fds)

        return wrapped_popen


class BubblewrapDriver(Driver, WrapperInterface):
    log = logging.getLogger("zuul.BubblewrapDriver")
    name = 'bubblewrap'

    release_file_re = re.compile(r'^\W+-release$')
    bwrap_version_re = re.compile(r'^(\d+\.\d+\.\d+).*')

    def __init__(self, check_bwrap):
        pid = os.getpid()
        with open(f"/proc/{pid}/oom_score_adj") as f:
            starting_score_adj = int(f.read().strip())

        self.oom_score_adj = min(1000, (starting_score_adj) + OOM_SCORE_ADJ)
        self.log.debug("Initializing bubblewrap with oom_score_adj "
                       "starting: %s, final: %s",
                       starting_score_adj, self.oom_score_adj)
        self.userns_enabled = self._is_userns_enabled()
        self.bwrap_version = self._parse_bwrap_version()
        self.bwrap_command = self._bwrap_command()
        if check_bwrap:
            # Validate basic bwrap functionality before we attempt to run
            # workloads under bwrap.
            context = self.getExecutionContext()
            popen = context.getPopen(work_dir='/tmp',
                                     ssh_auth_sock='/dev/null')
            p = popen(['id'],
                      stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            p.communicate()
            if p.returncode != 0:
                self.log.error("Non zero return code executing: %s",
                               " ".join(shlex.quote(c)
                                        for c in popen.command + ['id']))
                raise Exception('bwrap execution validation failed. You can '
                                'use `zuul-bwrap /tmp id` to investigate '
                                'manually.')

    def _is_userns_enabled(self):
        # This is based on the bwrap checks found here:
        # https://github.com/containers/bubblewrap/blob/
        # ad76c2d6ba8091a7afa95568e46af2261b362439/bubblewrap.c#L2735
        return_val = False
        if os.path.exists('/proc/self/ns/user'):
            # Rhel 7 specific case
            if os.path.exists('/sys/module/user_namespace/parameters/enable'):
                with open('/sys/module/user_namespace/parameters/enable') as f:
                    s = f.read()
                    if not s or s[0] != 'Y':
                        return return_val
            if os.path.exists('/proc/sys/user/max_user_namespaces'):
                with open('/proc/sys/user/max_user_namespaces') as f:
                    s = f.read()
                    try:
                        i = int(s.strip())
                        if i < 1:
                            return return_val
                    except ValueError:
                        # If we can't determine the max namespace count but
                        # namespaces are generally enabled then we should
                        # treat them as enabled.
                        return_val = True
            return_val = True
        return return_val

    def _parse_bwrap_version(self):
        p = subprocess.run(['bwrap', '--version'], text=True,
                           stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        # We don't know for sure what version schema bwrap may end up using.
        # Match what they have done historically and be a bit forgiving of
        # alpha, beta, rc, etc annotations.
        r = self.bwrap_version_re.match(p.stdout.split()[-1])
        if p.returncode == 0 and r:
            return tuple(map(int, r.group(1).split('.')))
        else:
            if p.returncode == 0:
                self.log.warning('Unable to determine bwrap version, from '
                                 '"%s". Using 0.0.0' % p.stdout.strip())
            else:
                self.log.warning('Unable to determine bwrap version, got '
                                 'returncode "%s". Using 0.0.0' % p.returncode)
            return (0, 0, 0)

    def reconfigure(self, tenant):
        pass

    def stop(self):
        pass

    def _bwrap_command(self):
        if self.bwrap_version >= (0, 8, 0) and self.userns_enabled:
            userns_flags = ['--unshare-user', '--disable-userns']
        else:
            userns_flags = []
        bwrap_command = [
            'setpriv',
            '--ambient-caps',
            '-all',
            'choom',
            '-n', str(self.oom_score_adj),
            '--',
            'bwrap',
            '--dir', '/tmp',
            '--tmpfs', '/tmp',
            '--dir', '/var',
            '--dir', '/var/tmp',
            '--dir', '/run/user/{uid}',
            '--ro-bind', '/usr', '/usr',
            '--ro-bind', '/lib', '/lib',
            '--ro-bind', '/bin', '/bin',
            '--ro-bind', '/sbin', '/sbin',
            '--ro-bind', '/etc/ld.so.cache', '/etc/ld.so.cache',
            '--ro-bind', '/etc/resolv.conf', '/etc/resolv.conf',
            '--ro-bind', '/etc/hosts', '/etc/hosts',
            '--ro-bind', '/etc/localtime', '/etc/localtime',
            '--ro-bind', '{ssh_auth_sock}', '{ssh_auth_sock}',
            '--bind', '{work_dir}', '{work_dir}',
            '--tmpfs', '{work_dir}/tmp',
            '--proc', '/proc',
            '--dev', '/dev',
            '--chdir', '{work_dir}',
            '--unshare-all',
            '--share-net',
            '--die-with-parent',
            '--uid', '{uid}',
            '--gid', '{gid}',
            '--file', '{uid_fd}', '/etc/passwd',
            '--file', '{gid_fd}', '/etc/group',
        ] + userns_flags

        for path in ['/lib64',
                     '/etc/nsswitch.conf',
                     '/etc/lsb-release.d',
                     '/etc/alternatives',
                     '/etc/ssl/certs',
                     '/etc/subuid',
                     '/etc/containers',
                     ]:
            if os.path.exists(path):
                bwrap_command.extend(['--ro-bind', path, path])
        for fn in os.listdir('/etc'):
            if self.release_file_re.match(fn):
                path = os.path.join('/etc', fn)
                bwrap_command.extend(['--ro-bind', path, path])

        return bwrap_command

    def getExecutionContext(self, ro_paths=None, rw_paths=None, secrets=None):
        if not ro_paths:
            ro_paths = []
        if not rw_paths:
            rw_paths = []
        if not secrets:
            secrets = {}
        return BubblewrapExecutionContext(
            self.bwrap_command,
            ro_paths, rw_paths,
            secrets)


def main(args=None):
    logging.basicConfig(level=logging.DEBUG)

    driver = BubblewrapDriver(check_bwrap=True)

    parser = argparse.ArgumentParser()
    parser.add_argument('--ro-paths', nargs='+')
    parser.add_argument('--rw-paths', nargs='+')
    parser.add_argument('--secret', nargs='+')
    parser.add_argument('work_dir')
    parser.add_argument('run_args', nargs='+')
    cli_args = parser.parse_args()

    # The zuul-bwrap command is often run for debugging purposes. An SSH
    # agent may not be necessary or present in that situation.
    ssh_auth_sock = os.environ.get('SSH_AUTH_SOCK', '/dev/null')

    secrets = {}
    if cli_args.secret:
        for secret in cli_args.secret:
            fn, content = secret.split('=', 1)
            secrets[fn] = content

    context = driver.getExecutionContext(
        cli_args.ro_paths, cli_args.rw_paths,
        secrets)

    popen = context.getPopen(work_dir=cli_args.work_dir,
                             ssh_auth_sock=ssh_auth_sock)
    x = popen(cli_args.run_args)
    x.wait()


if __name__ == '__main__':
    main()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Add QueueItem UUID column

Revision ID: 8e35d000c786
Revises: 6c1582c1d08c
Create Date: 2025-11-24 10:17:28.428121

"""

# revision identifiers, used by Alembic.
revision = '8e35d000c786'
down_revision = 'ce1459953a12'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade(table_prefix=''):
    prefixed_buildset_table = table_prefix + 'zuul_buildset'
    op.add_column(
        prefixed_buildset_table,
        sa.Column('queue_item_uuid', sa.String(36), nullable=True)
    )

    op.create_index(
        f'{prefixed_buildset_table}_queue_item_uuid_idx',
        prefixed_buildset_table,
        ['queue_item_uuid']
    )


def downgrade():
    raise Exception("Downgrades not supported")
